# -*- coding: utf-8 -*-
# Copyright (C) 2012-2017, Almar Klein
# This code is distributed under the terms of the (new) BSD License.

"""
This package provides two functions:
    * isocontour - for 2D images
    * isosurface - for 3D images

Visvis uses to implement this functionality itself using Cython. Visvis
was often used as a pure Python package though, so this functionality
was sort of hidden.

Therefore I've moved the marching cubes (isosurface) implementation
to scikit-image, and we now use that here.

"""

import visvis as vv
import numpy as np
from skimage.measure import find_contours, marching_cubes_lewiner


def isocontour(im, isovalue=None):
    """ isocontour(im, isovalue=None)
    
    Uses scikit-image to calculate the iso contours for the given 2D
    image. If isovalue is not given or None, a value between the min
    and max of the image is used.
    
    Returns a pointset in which each two subsequent points form a line
    piece. This van be best visualized using "vv.plot(result, ls='+')".
    
    """
    
    # Check image
    if not isinstance(im, np.ndarray) or (im.ndim != 2):
        raise ValueError('im should be a 2D numpy array.')
    
    # Get isovalue
    if isovalue is None:
        isovalue = 0.5 * (im.min() + im.max())
    isovalue = float(isovalue) # Will raise error if not float-like value given
    
    # Get the contours
    data = find_contours(im, isovalue)
    
    # Build the contour as we used to return it. It is less rich, but easier
    # to visualize.
    data2 = []
    for contour in data:
        n = contour.shape[0] * 2 - 2
        contour2 = np.empty((n, 2), np.float32)
        contour2[0::2] = contour[:-1]
        contour2[1::2] = contour[1:]
        data2.append(np.fliplr(contour2))
    
    # Return as pointset
    return vv.Pointset(np.row_stack(data2))


def isosurface(im, isovalue=None, step=1, useClassic=False, useValues=False):
    """ isosurface(vol, isovalue=None, step=1, useClassic=False, useValues=False)
    
    Uses scikit-image to calculate the isosurface for the given 3D image.
    Returns a vv.BaseMesh object.
    
    Parameters
    ----------
    vol : 3D numpy array
        The volume for which to calculate the isosurface.
    isovalue : float
        The value at which the surface should be created. If not given or None,
        the average of the min and max of vol is used.
    step : int
        The stepsize for stepping through the volume. Larger steps yield
        faster but coarser results. The result shall always be topologically
        correct though.
    useClassic : bool
        If True, uses the classic marching cubes by Lorensen (1987) is used.
        This algorithm has many ambiguities and is not guaranteed to produce
        a topologically correct result.
    useValues : bool
        If True, the returned BaseMesh object will also have a value for
        each vertex, which is related to the maximum value in a local region
        near the isosurface.
    
    """
    
    # Check image
    if not isinstance(im, np.ndarray) or (im.ndim != 3):
        raise ValueError('vol should be a 3D numpy array.')
    
    # Get isovalue
    if isovalue is None:
        isovalue = 0.5 * (im.min() + im.max())
    isovalue = float(isovalue) # Will raise error if not float-like value given
    
    # Check steps
    step = int(step)
    if step < 1:
        raise ValueError('step must be at least one.')
    
    # Deal with Aarray
    if hasattr(im, 'sampling'):
        sampling = im.sampling[0], im.sampling[1], im.sampling[2]
    else:
        sampling = (1, 1, 1)
    
    # Call into skimage algorithm
    xx = marching_cubes_lewiner(im, isovalue,
                                spacing=sampling,
                                step_size=step,
                                use_classic=useClassic)
    vertices, faces, normals, values = xx
    
    # Transform the data to how Visvis likes it
    vertices = np.fliplr(vertices)
    
    # Check
    if not len(vertices):
        raise RuntimeError('No surface found at the given iso value.')
    
    # Done
    if useValues:
        return vv.BaseMesh(vertices, faces, normals, values)
    else:
        return vv.BaseMesh(vertices, faces, normals)



#!/usr/bin/env python
""" Examples of using qhull via scipy to generate 3D plots in visvis.
Requires numpy ver 1.5, scipy ver 0.9 and qhull from
http://www.qhull.org/ (on Windows this comes with Scipy).

plot3D meshes and plots random convex transformable data in both cartesian
and spherical coordinates

Play around with the many input parameters to plot3D to make interesting plots.

Keith Smith, 4 March 2011

"""

import numpy as np
from scipy.spatial import Delaunay
import visvis as vv


def plot3D( vuvi,
            coordSys='Cartesian',
            raised = True,
            depRange=[-40,0],
            ambient = 0.9,
            diffuse = 0.4,
            colormap = vv.CM_JET,
            faceShading='smooth',
            edgeColor = (0.5,0.5,0.5,1),
            edgeShading = 'smooth',
            faceColor = (1,1,1,1),
            shininess = 50,
            specular = 0.35,
            emission = 0.45 ):
    """ plot3D(vxyz,
            coordSys=['Cartesian', 'Spherical'],
            raised = True,
            depRange=[-40,0], #Note: second range limit not currently used
            rangeR=[-40,0],
            ambient = 0.9,
            diffuse = 0.4,
            colormap = vv.CM_JET,
            faceShading='smooth',
            edgeColor = (0.5,0.5,0.5,1),
            edgeShading = 'smooth',
            faceColor = (1,1,1,1),
            shininess = 50,
            specular = 0.35,
            emission = 0.45 ))
    """
  
    if coordSys == 'Spherical':
        thetaPhiR = vuvi  # data cols are theta, phi, radius
        vxyz = np.zeros(vuvi.shape)
        
        # Now find xyz data points on unit sphere (for meshing)
        vxyz[:,0] = np.sin(thetaPhiR[:,0])*np.cos(thetaPhiR[:,1])
        vxyz[:,1] = np.sin(thetaPhiR[:,0])*np.sin(thetaPhiR[:,1])
        vxyz[:,2] = np.cos(thetaPhiR[:,0])
        
        #normalize and scale dependent values
        thetaPhiR[thetaPhiR[:,2] < depRange[0], 2] = depRange[0]
        depVal = thetaPhiR[:,2] - np.min(thetaPhiR[:,2])
    else:
        vxyz = vuvi
        vxyz[vxyz[:,2] < depRange[0], 2] = depRange[0]
        numOfPts = np.shape(vxyz)[0]
        depVal = vxyz[:,2]
        # set to convex surface for meshing
        # find center of data
        center = np.average(vxyz, 0)
        #center data
        vxyz = vxyz - center
        # find x-y plane distance to each point
        radials = np.sqrt(vxyz[:,0]**2 + vxyz[:,1]**2)
        # get max and adjust so that arctan ranges between +-45 deg
        maxRadial = np.max(radials)/0.7
        #get angle on sphere
        xi =  np.arctan2(radials / maxRadial, 1)
        #force z axis data to sphere
        vxyz[:,2] = maxRadial * np.cos(xi)
        vxyz = np.append(vxyz, [[0.7, 0.7, -0.7],[-0.7, 0.7, -0.7],
                                [0.7, -0.7, -0.7],[-0.7, -0.7, -0.7]], axis=0)
       
    
    
    # Send data to convex_hull program qhull
    dly = Delaunay(vxyz)
    meshIndx = dly.convex_hull
    
    # Check each triangle facet and flip if
    # vertex order puts back side out
    for index, (I1, I2, I3) in enumerate(meshIndx):
        a = vxyz[I1,:] - vxyz[I2,:]
        b = vxyz[I2,:] - vxyz[I3,:]
        c = np.cross(a, b)
        if np.dot(c, vxyz[I2,:]) > 0:
            meshIndx[index] = (I1, I3, I2)

    
    
    # if 3D surface adjust dependent coordinates
    if raised:
        if coordSys == 'Spherical':
            vxyz[:,0] = depVal*np.sin(thetaPhiR[:,0])*np.cos(thetaPhiR[:,1])
            vxyz[:,1] = depVal*np.sin(thetaPhiR[:,0])*np.sin(thetaPhiR[:,1])
            vxyz[:,2] = depVal*np.cos(thetaPhiR[:,0])
        else:
            vxyz = vxyz + center
            vxyz[:numOfPts,2] = depVal
    else:
        
        if coordSys == 'Spherical':
            depRange[0] = 1.0
        else:
            # Since qhull encloses the data with Delaunay triangles there will be
            # a set of facets which cover the bottom of the data.  For flat
            # contours, the bottom facets need to be separated a fraction from
            # the top facets else you don't see colormap colors
            depValRange = np.max(vxyz[:numOfPts,2]) - np.min(vxyz[:numOfPts,2])
            vxyz[:numOfPts,2] = vxyz[:numOfPts,2] /   (10 * depValRange )
            
    #normalize depVal for color mapping
    dataRange = np.max(depVal) - np.min(depVal)
    depVal = (depVal- np.min(depVal)) /  dataRange

    # Get axes
    ax = vv.gca()
    
    ms = vv.Mesh(ax, vxyz, faces=meshIndx, normals=vxyz)
    ms.SetValues(np.reshape(depVal,np.size(depVal)))
    ms.ambient = ambient
    ms.diffuse = diffuse
    ms.colormap = colormap
    ms.faceShading = faceShading
    ms.edgeColor = edgeColor
    ms.edgeShading = edgeShading
    ms.faceColor = faceColor
    ms.shininess = shininess
    ms.specular = specular
    ms.emission = emission
    ax.SetLimits(rangeX=[-depRange[0],depRange[0]],
                 rangeY=[-depRange[0],depRange[0]],
                 rangeZ=[-depRange[0], depRange[0]])



# Start of test code.
if __name__ == '__main__':
    
    # Create figure
    fig = vv.figure()
    fig.position.w = 600
    
    
    # Cartesian plot
    numOfPts = 2000
    scale = 1
    
    # Create random points
    xyz = 2 * scale * (np.random.rand(numOfPts,3) - 0.5)
    
    # 2D sync function
    xyz[:,2] =  np.sinc(5*(np.sqrt(xyz[:,0]**2 + xyz[:,1]**2)))
    #xyz[:,2] =  scale - ( xyz[:,0]**2 + xyz[:,1]**2)
    
    # Plot
    vv.subplot(121)
    vv.title('Cartesian coordinates')
    plot3D(xyz, depRange=[-1,0])
    #plot3D(xyz, depRange=[-1,0], raised=False)
    
    
    # Sperical plot
    numOfPts = 1000
    
    # Create random points
    ThetaPhiR = np.zeros((numOfPts,3))
    ThetaPhiR[:,0] = np.pi * np.random.rand(numOfPts) # theta is 0 to 180 deg
    ThetaPhiR[:,1] = 2 * np.pi * np.random.rand(numOfPts) # phi is 0 to 360 deg
    ThetaPhiR[:,2] = 10 * np.log10((np.sin(ThetaPhiR[:,0])**4) * (np.cos(ThetaPhiR[:,1])**2))
    
    # Plot
    vv.subplot(122)
    vv.title('Sperical coordinates')
    plot3D(ThetaPhiR, coordSys='Spherical')
    #plot3D(ThetaPhiR, coordSys='Spherical', raised=False)
    
    
    # Run main loop
    app = vv.use()
    app.Run()



# -*- coding: utf-8 -*-
# Copyright (C) 2012, Almar Klein
#
# Visvis is distributed under the terms of the (new) BSD License.
# The full license can be found in 'license.txt'.

import visvis as vv
from visvis import Axes, AxesContainer



def getcenter(f, a):
    """ Get the center (in relative coords) of the axes)."""
    pos = a.position.Copy()
    # make relative
    relative = pos._GetFractionals()
    for i in range(4):
        if not relative[i]:
            pos[i] = pos[i] / f.position[i%2+2]
        if pos[i]<0:
            pos[i] = 1.0 - pos[i]
    return pos.x + pos.w/2.0, pos.y + pos.h/2.0


def laysin(cols, rows, c):
    """ Given the center and the amount of rows and columns,
    return in which bin the axes is in.
    """
    dx, dy = 1.0/cols, 1.0/rows
    x, y = int(c[0]/dx), int(c[1]/dy)
    return x + y*cols
    

def subplot(*args):
    """ subplot(ncols, nrows, nr)
    
    Create or return axes in current figure. Note that subplot(322) is the
    same as subplot(3,2,2).
    
    Parameters
    ----------
    ncols : int
        The number of columns to devide the figure in.
    nrows : int
        The number of rows to devide the figure in.
    nr : int
        The subfigure number on the grid specified by ncols and nrows.
        Should be at least one. subplot(221) is the top left. subplot(222)
        is the top right.
    
    Notes
    -----
    It is checked whether (the center of) an axes is present at the
    specified grid location. If so, that axes is returned. Otherwise
    a new axes is created at that location.
    
    """
    
    # parse input
    if len(args)==1:
        tmp = args[0]
        if tmp>999 or tmp<111:
            raise ValueError("Invalid cols/rows/nr specified.")
        rows = tmp // 100
        tmp  = tmp % 100
        cols = tmp // 10
        tmp  = tmp % 10
        nr   = tmp
    elif len(args)==3:
        rows, cols, nr = args
    else:
        raise ValueError("Invalid number of cols/rows/nr specified.")
    
    # check if ok
    if nr<=0 or cols<=0 or rows<=0:
        raise ValueError("Invalid cols/rows/nr: all bust be >0.")
    if nr > cols*rows:
        raise ValueError("Invalid nr: there are not so many positions.")
        
    # init
    f = vv.gcf()
    nr = nr-1
    
    # check if an axes is there
    for a in f._children:
        if isinstance(a, AxesContainer):
            n = laysin( cols, rows, getcenter(f, a) )
            if n == nr:
                # make current and return
                a = a.GetAxes()
                f.currentAxes = a
                return a
    
    # create axes in container
    a = Axes(f)
    c = a.parent
    
    # calculate relative coordinates
    dx, dy = 1.0/cols, 1.0/rows
    y = int( nr / cols )
    x = int( nr % cols )
    
    # apply positions
    c.position = dx*x, dy*y, dx, dy
    a.position = 10, 10, -20, -20
    
    # done
    return a
    
    

if __name__ == "__main__":
    f = vv.figure()
    # Get axes on 2x2 grid
    a1 = vv.subplot(221)
    a2 = vv.subplot(221)
    a3 = vv.subplot(224)
    # Get axes on 3x3 grid
    a4 = vv.subplot(331) # (center is inside 221)
    a5 = vv.subplot(333)
    # Check
    print(a1 is a2, a1 is a4)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

""" This example illustrate using text and formatting for text
world objects and labels.
"""

import visvis as vv

# Define a piece of Unicode text in a py2.x py3.x compatible manner
hello = u'Привет пустошь'

# Create figure and figure
fig = vv.figure()
a = vv.cla()
a.cameraType = '2d'
a.daspectAuto = True
a.SetLimits((0,8), (-1,10))

# Create text inside the axes
t1 = vv.Text(a, 'Visvis text', 0.2, 9, 0, 'mono', 30)
t2 = vv.Text(a, '... with FreeType!', 1, 8, 0, 'serif', 12)
t3 = vv.Text(a, '... and Unicode: %s!' % hello, 1, 7)
t3 = vv.Text(a, '\Gamma\rho\epsilon\epsilon\kappa letters and ' +
                ' \rightarrow math \otimes symbols', 1, 6)

t2 = vv.Text(a, '\b{bold}, \i{italic}, and \b{\i{bolditalic}} \bfon\its' +
                ' and sup^{script} and sub_{script}', 1, 5, 0, 'serif')
t3 = vv.Text(a, 'Look, I\'m at an angle!', 1, 4)
t3.textAngle = -20
t3.fontSize = 12


# Create text labels
label1 = vv.Label(a, 'This is a Label')
label1.position = 10, 0.9
label1.bgcolor = (0.5, 1, 0.5)

## A quick brown fox jumps over the lazy dog

testText = 'A quick brown fox jumps over the lazy dog'

# Create figure and figure
fig = vv.figure()
a = vv.cla()
a.cameraType = '2d'

vv.title(testText)
# Create text labels
for i in range(20):
    offset = 0.1 * i
    label = vv.Label(a, '| ' + testText + ' - %1.2f pixels offset' % offset)
    label.bgcolor = None
    label.position = 10 + 0.1*i, 10 + i*15


# Enter main loop
app = vv.use()
app.Run()



#!/usr/bin/env python

import numpy as np
import visvis as vv

# Read image and crop
astro = vv.imread('astronaut.png').astype(np.float32)
astro = astro[100:-100,100:-100, :]

# Smooth a bit
im = astro.copy()
im[1:,:,:] = astro[:-1,:,:]
im[:-1,:,:] += astro[1:,:,:]
im[:,:-1,:] += astro[:,1:,:]
im[:,1:,:] += astro[:,:-1,:]
im /= 4

# Prepare figure
vv.figure()

# Without color, with colormap
a1 = vv.subplot(121)
m1 = vv.surf(im[:,:,0])
m1.colormap = vv.CM_HOT
vv.title('With colormap')

# With color
a2 = vv.subplot(122)
m2 = vv.surf(im[:,:,0], im)
vv.title('With original texture')

# Flip y-axis, otherwise the image is upside down
a1.daspect = 1,-1,1
a2.daspect = 1,-1,1

# Enter mainloop
app=vv.use()
app.Run()



#!/usr/bin/env python
"""
This example illustrates the use of the Aarray class.
Particularly how the class is aware of slicing and adapts
the origin and sampling accordingly. Visvis uses the sampling
and origin of the Aarray instances to draw the image at the
correct scale and position.
"""

import visvis as vv
from visvis import Aarray
app = vv.use()

# Load image and make Aarray
im = vv.imread('astronaut.png')
im = Aarray(im)

# Cut in four pieces, but also change resolution using different step sizes
im1 = im[:300,:300]
im2 = im[300:,:300:7]
im3 = im[:300:5,300:]
im4 = im[300::4,300::4]

# Get an axes
a = vv.gca()

# Show all images
tt = []
for im in [im1, im2, im3, im4]:
    tt.append(vv.imshow(im))

# Note that some parts seem to stick out. This is because visvis
# renders data such that the pixel center is at the spefied position;
# larger pixels thus stick out more.

# Enter mainloop
app.Run()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
#
#  FreeType high-level python API - Copyright 2011 Nicolas P. Rougier
#  Distributed under the terms of the new BSD license.
#
# -----------------------------------------------------------------------------
'''
Freetype basic data types
-------------------------

FT_Byte : A simple typedef for the unsigned char type.

FT_Bytes : A typedef for constant memory areas.

FT_Char : A simple typedef for the signed char type.

FT_Int : A typedef for the int type.

FT_UInt : A typedef for the unsigned int type.

FT_Int16 : A typedef for a 16bit signed integer type.

FT_UInt16 : A typedef for a 16bit unsigned integer type.

FT_Int32 : A typedef for a 32bit signed integer type.

FT_UInt32 : A typedef for a 32bit unsigned integer type.

FT_Short : A typedef for signed short.

FT_UShort : A typedef for unsigned short.

FT_Long : A typedef for signed long.

FT_ULong : A typedef for unsigned long.

FT_Bool : A typedef of unsigned char, used for simple booleans. As usual,
          values 1 and 0 represent true and false, respectively.

FT_Offset : This is equivalent to the ANSI C 'size_t' type, i.e., the largest
            unsigned integer type used to express a file size or position, or
            a memory block size.

FT_PtrDist : This is equivalent to the ANSI C 'ptrdiff_t' type, i.e., the
             largest signed integer type used to express the distance between
             two pointers.

FT_String : A simple typedef for the char type, usually used for strings.

FT_Tag  : A typedef for 32-bit tags (as used in the SFNT format).

FT_Error : The FreeType error code type. A value of 0 is always interpreted as
           a successful operation.

FT_Fixed : This type is used to store 16.16 fixed float values, like scaling
           values or matrix coefficients.

FT_Pointer : A simple typedef for a typeless pointer.

FT_Pos : The type FT_Pos is used to store vectorial coordinates. Depending on
         the context, these can represent distances in integer font units, or
         16.16, or 26.6 fixed float pixel coordinates.

FT_FWord : A signed 16-bit integer used to store a distance in original font
           units.

FT_UFWord : An unsigned 16-bit integer used to store a distance in original
            font units.

FT_F2Dot14 : A signed 2.14 fixed float type used for unit vectors.

FT_F26Dot6 : A signed 26.6 fixed float type used for vectorial pixel
             coordinates.
'''
from ctypes import (CFUNCTYPE, c_void_p, c_size_t,
                    c_ubyte, c_char, c_char_p,
                    c_short, c_ushort, c_int, c_uint, c_int32, c_uint32,
                    c_long, c_ulong, c_longlong)


FT_Byte    = c_ubyte  # A simple typedef for the unsigned char type.

FT_Bytes   = c_char_p # A typedef for constant memory areas.

FT_Char    = c_char   # A simple typedef for the signed char type.

FT_Int     = c_int    # A typedef for the int type.

FT_UInt    = c_uint   # A typedef for the unsigned int type.

FT_Int16   = c_short  # A typedef for a 16bit signed integer type.

FT_UInt16  = c_ushort # A typedef for a 16bit unsigned integer type.

FT_Int32   = c_int32  # A typedef for a 32bit signed integer type.

FT_UInt32  = c_uint32 # A typedef for a 32bit unsigned integer type.

FT_Short   = c_short  # A typedef for signed short.

FT_UShort  = c_ushort # A typedef for unsigned short.

FT_Long    = c_long   # A typedef for signed long.

FT_ULong   = c_ulong  # A typedef for unsigned long.

FT_Bool    = c_char   # A typedef of unsigned char, used for simple booleans. As
                      # usual, values 1 and 0 represent true and false,
                      # respectively.

FT_Offset  = c_size_t # This is equivalent to the ANSI C 'size_t' type, i.e.,
                      # the largest unsigned integer type used to express a file
                      # size or position, or a memory block size.

FT_PtrDist = c_longlong # This is equivalent to the ANSI C 'ptrdiff_t' type,
                        # i.e., the largest signed integer type used to express
                        # the distance between two pointers.

FT_String  = c_char   # A simple typedef for the char type, usually used for strings.

FT_String_p= c_char_p

FT_Tag     = FT_UInt32 # A typedef for 32-bit tags (as used in the SFNT format).

FT_Error   = c_int    # The FreeType error code type. A value of 0 is always
                      # interpreted as a successful operation.

FT_Fixed   = c_long   # This type is used to store 16.16 fixed float values,
                      # like scaling values or matrix coefficients.

FT_Pointer = c_void_p # A simple typedef for a typeless pointer.

FT_Pos     = c_long   # The type FT_Pos is used to store vectorial
                      # coordinates. Depending on the context, these can
                      # represent distances in integer font units, or 16.16, or
                      # 26.6 fixed float pixel coordinates.

FT_FWord   = c_short  # A signed 16-bit integer used to store a distance in
                      # original font units.

FT_UFWord  = c_ushort # An unsigned 16-bit integer used to store a distance in
                      # original font units.

FT_F2Dot14 = c_short  # A signed 2.14 fixed float type used for unit vectors.

FT_F26Dot6 = c_long   # A signed 26.6 fixed float type used for vectorial pixel
                      # coordinates.

FT_Glyph_Format = c_int

FT_Encoding     = c_int


# Describe a function used to destroy the 'client' data of any FreeType
# object. See the description of the FT_Generic type for details of usage.
FT_Generic_Finalizer = CFUNCTYPE(None, c_void_p)



#!/usr/bin/env python
"""
This example illustrates embedding a visvis figure in an FLTK application.
"""

import fltk
import visvis as vv

# Create a visvis app instance, which wraps an fltk application object.
# This needs to be done *before* instantiating the main window.
app = vv.use('fltk')


class MainWindow(fltk.Fl_Window):
    def __init__(self):
        fltk.Fl_Window.__init__(self, 560, 420, "Embedding in FLTK")
        
        # Make a panel with a button
        but = fltk.Fl_Button(10,10,70,30, 'Click me')
        but.callback(self._Plot)
        
        # Make figure to draw stuff in
        Figure = app.GetFigureClass()
        self.fig = Figure(100,10,560-110,420-20, "")
        
        # Make box for resizing
        box = fltk.Fl_Box(fltk.FL_NO_BOX,100,50, 560-110,420-60,"")
        self.resizable(box)
        box.hide()
        
        # Finish
        self.end()
        self.show()
        self.fig._widget.show()
    
    
    def _Plot(self, event):
        
        # Make sure our figure is the active one
        # If only one figure, this is not necessary.
        #vv.figure(self.fig.nr)
        
        # Clear it
        vv.clf()
        
        # Plot
        vv.plot([1,2,3,1,6])
        vv.legend(['this is a line'])


# Two ways to create the application and start the main loop
if True:
    # The visvis way. Will run in interactive mode when used in IEP or IPython.
    app.Create()
    m = MainWindow()
    app.Run()

else:
    # The native way.
    m = MainWindow()
    fltk.Fl.run()



# -*- coding: utf-8 -*-
# Copyright (C) 2012, Almar Klein
#
# Visvis is distributed under the terms of the (new) BSD License.
# The full license can be found in 'license.txt'.

import visvis as vv
import os


# Try importing imageio
imageio = None
try:
    import imageio
except ImportError:
    pass


def volread(filename):
    """ volread(filename)
    
    Read volume from a file. If filename is 'stent', read a dedicated
    test dataset. For reading any other kind of volume, the imageio
    package is required.
    
    """
    
    # Try loading our base volume(s)
    if filename == 'stent':
        path = vv.misc.getResourceDir()
        filename2 = os.path.join(path, 'stent_vol.ssdf')
        if os.path.isfile(filename2):
            filename = filename2
            s = vv.ssdf.load(filename)
            return s.vol.astype('int16') * s.colorscale
    
    # Use imageio (can also load from http, etc)
    if imageio is None:
        raise RuntimeError("visvis.volread needs the imageio package to read arbitrary files.")
    return imageio.volread(filename)


if __name__ == '__main__':
    vol = vv.volread('stent')
    t = vv.volshow(vol)
    t.renderStyle = 'mip' # maximum intensity projection (is the default)



# -*- coding: utf-8 -*-
# Copyright (C) 2012, Almar Klein
#
# Visvis is distributed under the terms of the (new) BSD License.
# The full license can be found in 'license.txt'.

from visvis import Label

class Title(Label):
    """ Title(axes, text)
    
    A label wibject that places itself at the top of its parent (which
    should be an Axes). The axes' position is corrected so the title
    will fit on screen.
    
    A Title can be created with the function vv.title().
    
    """
    
    def __init__(self, axes, text):
        Label.__init__(self, axes, text)
        
        # set textsize and align
        self.halign = 0
        self.fontSize = 12
        
        # set back color to be transparant
        self.bgcolor = ''
        
        # set position
        self.position = 0, -20, 1, 15
        
        # correct axes' position
        dy = -20
        axes.position.Correct(0, -dy, 0, dy)
    
    
    def OnDestroy(self):
        Label.OnDestroy(self)
        
        # correct axes' position
        axes = self.parent
        if axes:
            dy = 20
            axes.position.Correct(0, -dy, 0, dy)



#!/usr/bin/env python
""" This example illustrates how you can visualize
a volume using 2D slices.

We explicitly use volshow2() here. Note that volshow() calls
volshow3() by default, but volshow2() when the system does
not support 3D rendering (OpenGL version <2.0).

"""


import visvis as vv
vol = vv.aVolume()

t = vv.volshow2(vol)
t.colormap = vv.CM_HOT
t.edgeColor = (0.8, 0.8, 0.8)
app = vv.use()
app.Run()



# -*- coding: utf-8 -*-
# Copyright (C) 2012, Almar Klein
#
# Visvis is distributed under the terms of the (new) BSD License.
# The full license can be found in 'license.txt'.

""" Module sliceTextures

Defines texture wobjects for visualizing slices in 3D volumes.

"""

import OpenGL.GL as gl
import OpenGL.GLU as glu

from visvis.utils.pypoints import Point
from visvis import Wobject, Colormapable
from visvis.core.misc import Property, PropWithDraw, getColor
from visvis.core import shaders
from visvis.wobjects.textures import BaseTexture, TextureObjectToVisualize


class SliceTexture(BaseTexture):
    """ SliceTexture
    
    A slice texture is a 2D texture of a 3D data volume. It enables
    visualizing 3D data without the need for glsl renderering (and can
    therefore be used on older systems.
    
    """
    
    def __init__(self, parent, data, axis=0, index=0):
        BaseTexture.__init__(self, parent, data)
        self._ndim = 3
        
        # Init parameters
        self._axis = axis
        self._index = index
        
        # create texture
        self._texture1 = TextureObjectToVisualize(2, data)
        
        # init shader
        self._InitShader()
        
        # set data  (data to textureToV. only for min/max)
        self.SetData(data)
        
        # init interpolation
        self._texture1._interpolate = True
        
        # For edge
        self._edgeColor = None
        self._edgeColor2 = getColor('g')
        self._edgeWidth = 3.0
        
        # For interaction
        self._interact_over = False
        self._interact_down = False
        self._screenVec = None
        self._refPos = (0,0)
        self._refIndex = 0
        #
        self.eventEnter.Bind(self._OnMouseEnter)
        self.eventLeave.Bind(self._OnMouseLeave)
        self.eventMouseDown.Bind(self._OnMouseDown)
        self.eventMouseUp.Bind(self._OnMouseUp)
        self.eventMotion.Bind(self._OnMouseMotion)
    
    
    def _InitShader(self):
        
        # Add components of shaders
        self.shader.vertex.Clear()
        self.shader.fragment.Clear()
        self.shader.fragment.AddPart(shaders.SH_2F_BASE)
        self.shader.fragment.AddPart(shaders.SH_2F_AASTEPS_0)
        self.shader.fragment.AddPart(shaders.SH_COLOR_SCALAR)
        
        def uniform_shape():
            shape = self._texture1._shape[:2] # as in opengl
            return [float(s) for s in reversed(list(shape))]
        def uniform_extent():
            data = self._texture1._dataRef # as original array
            shape = reversed(data.shape[:2])
            if hasattr(data, 'sampling'):
                sampling = reversed(data.sampling[:2])
            else:
                sampling = [1.0 for s in range(2)]
            
            del data
            return [s1*s2 for s1, s2 in zip(shape, sampling)]
        
        # Set some uniforms
        self.shader.SetStaticUniform('colormap', self._colormap)
        self.shader.SetStaticUniform('shape', uniform_shape)
        self.shader.SetStaticUniform('scaleBias', self._texture1._ScaleBias_get)
        self.shader.SetStaticUniform('extent', uniform_extent)
        self.shader.SetStaticUniform('aakernel', [1.0, 0, 0, 0])
    
    
    def _SetData(self, data):
        """ _SetData(data)
        
        Give reference to the raw data. For internal use. Inheriting
        classes can override this to store data in their own way and
        update the OpenGL textures accordingly.
        
        """
        
        # Store data
        self._dataRef3D = data
        
        # Slice it
        i = self._index
        if self._axis == 0:
            slice = self._dataRef3D[i]
        elif self._axis == 1:
            slice = self._dataRef3D[:,i]
        elif self._axis == 2:
            slice = self._dataRef3D[:,:,i]
        
        # Update texture
        self._texture1.SetData(slice)
    
    
    def _GetData(self):
        """ _GetData()
        
        Get a reference to the raw data. For internal use.
        
        """
        return self._dataRef3D
    
    
    def _GetLimits(self):
        """ Get the limits in world coordinates between which the object exists.
        """
        
        # Obtain untransformed coords
        shape = self._dataRef3D.shape
        x1, x2 = -0.5, shape[2]-0.5
        y1, y2 = -0.5, shape[1]-0.5
        z1, z2 = -0.5, shape[0]-0.5
        
        # There we are
        return Wobject._GetLimits(self, x1, x2, y1, y2, z1, z2)
    
    
    def OnDestroy(self):
        # Clear normaly, and also remove reference to data
        BaseTexture.OnDestroy(self)
        self._dataRef3D = None
    
    
    def OnDrawShape(self, clr):
        # Implementation of the OnDrawShape method.
        gl.glColor(clr[0], clr[1], clr[2], 1.0)
        self._DrawQuads()
    
    
    def OnDraw(self, fast=False):
        # Draw the texture.
        
        # set color to white, otherwise with no shading, there is odd scaling
        gl.glColor3f(1.0,1.0,1.0)
        
        # Enable texture, so that it has a corresponding OpenGl texture.
        # Binding is done by the shader
        self._texture1.Enable(-1)
        self.shader.SetUniform('texture', self._texture1)
        
        # _texture._shape is a good indicator of a valid texture
        if not self._texture1._shape:
            return
        
        if self.shader.isUsable and self.shader.hasCode:
            # fragment shader on -> anti-aliasing
            self.shader.Enable()
        else:
            # Fixed funcrion pipeline
            self.shader.EnableTextureOnly('texture')
        
        # do the drawing!
        self._DrawQuads()
        gl.glFlush()
        
        # clean up
        self.shader.Disable()
        
        
        # Draw outline?
        clr = self._edgeColor
        if self._interact_down or self._interact_over:
            clr = self._edgeColor2
        if clr:
            self._DrawQuads(clr)
        
        # Get screen vector?
        if self._screenVec is None:
            pos1 = [int(s/2) for s in self._dataRef3D.shape]
            pos2 = [s for s in pos1]
            pos2[self._axis] += 1
            #
            screen1 = glu.gluProject(pos1[2], pos1[1], pos1[0])
            screen2 = glu.gluProject(pos2[2], pos2[1], pos2[0])
            #
            self._screenVec = screen2[0]-screen1[0], screen1[1]-screen2[1]
    
    
    def _DrawQuads(self, clr=None):
        """ Draw the quads of the texture.
        This is done in a seperate method to reuse code in
        OnDraw() and OnDrawShape().
        """
        if not self._texture1._shape:
            return
        
        # The -0.5 offset is to center pixels/voxels. This works correctly
        # for anisotropic data.
        x1, x2 = -0.5, self._dataRef3D.shape[2]-0.5
        y2, y1 = -0.5, self._dataRef3D.shape[1]-0.5
        z2, z1 = -0.5, self._dataRef3D.shape[0]-0.5
        
        # Calculate quads
        i = self._index
        if self._axis == 0:
            quads = [   (x1, y2, i),
                        (x2, y2, i),
                        (x2, y1, i),
                        (x1, y1, i),    ]
        elif self._axis == 1:
            quads = [   (x1, i, z2),
                        (x2, i, z2),
                        (x2, i, z1),
                        (x1, i, z1),    ]
        elif self._axis == 2:
            quads = [   (i, y2, z2),
                        (i, y1, z2),
                        (i, y1, z1),
                        (i, y2, z1),    ]
        
        if clr:
            # Draw lines
            gl.glColor(clr[0], clr[1], clr[2], 1.0)
            gl.glLineWidth(self._edgeWidth)
            gl.glBegin(gl.GL_LINE_STRIP)
            for i in [0,1,2,3,0]:
                gl.glVertex3d(*quads[i])
            gl.glEnd()
        else:
            # Draw texture
            gl.glBegin(gl.GL_QUADS)
            gl.glTexCoord2f(0,0); gl.glVertex3d(*quads[0])
            gl.glTexCoord2f(1,0); gl.glVertex3d(*quads[1])
            gl.glTexCoord2f(1,1); gl.glVertex3d(*quads[2])
            gl.glTexCoord2f(0,1); gl.glVertex3d(*quads[3])
            gl.glEnd()
    
    
    ## Interaction
    
    def _OnMouseEnter(self, event):
        self._interact_over = True
        self.Draw()
    
    def _OnMouseLeave(self, event):
        self._interact_over = False
        self.Draw()
    
    def _OnMouseDown(self, event):
        
        if event.button == 1:
            
            # Signal that its down
            self._interact_down = True
            
            # Make the screen vector be calculated on the next draw
            self._screenVec = None
            
            # Store position and index for reference
            self._refPos = event.x, event.y
            self._refIndex = self._index
            
            # Redraw
            self.Draw()
            
            # Handle the event
            return True
        
        else:
            event.Ignore()
    
    def _OnMouseUp(self, event):
        self._interact_down = False
        self.Draw()
    
    def _OnMouseMotion(self, event):
        
        # Handle or pass?
        if not (self._interact_down and self._screenVec):
            return
        
        # Get vector relative to reference position
        refPos = Point(self._refPos)
        pos = Point(event.x, event.y)
        vec = pos - refPos
        
        # Length of reference vector, and its normalized version
        screenVec = Point(self._screenVec)
        L = screenVec.norm()
        V = screenVec.normalize()
        
        # Number of indexes to change
        n = vec.dot(V) / L
        
        # Apply!
        self.index = int(self._refIndex + n)
    
    
    ## Properties
    
    
    @PropWithDraw
    def index():
        """ The index of the slice in the volume to display.
        """
        def fget(self):
            return self._index
        def fset(self, value):
            # Check value
            if value < 0:
                value = 0
            maxIndex = self._dataRef3D.shape[self._axis] - 1
            if value > maxIndex:
                value = maxIndex
            # Set and update
            self._index = value
            self._SetData(self._dataRef3D)
        return locals()
    
    
    @PropWithDraw
    def axis():
        """ The axis of the slice in the volume to display.
        """
        def fget(self):
            return self._axis
        def fset(self, value):
            # Check value
            if value < 0 or value >= 3:
                raise ValueError('Invalid axis.')
            # Set and update index (can now be out of bounds.
            self._axis = value
            self.index = self.index
        return locals()
    
    
    @PropWithDraw
    def edgeColor():
        """ The color of the edge of the slice (can be None).
        """
        def fget(self):
            return self._edgeColor
        def fset(self, value):
            self._edgeColor = getColor(value)
        return locals()
    
    
    @PropWithDraw
    def edgeColor2():
        """ The color of the edge of the slice when interacting.
        """
        def fget(self):
            return self._edgeColor2
        def fset(self, value):
            self._edgeColor2 = getColor(value)
        return locals()
    

class SliceTextureProxy(Wobject, Colormapable):
    """ SliceTextureProxy(*sliceTextures)
    
    A proxi class for multiple SliceTexture instances. By making them
    children of an instance of this class, their properties can be
    changed simultaneously.
    
    This makes it possible to call volshow() and stay agnostic of how
    the volume is vizualized (using a 3D render, or with 3 slice
    textures); all public texture-specific methods and properties are
    transferred to all children automatically.
    
    """
    
    
    def SetData(self, *args, **kwargs):
        for s in self.children:
            s.SetData(*args, **kwargs)
    
    def Refresh(self, *args, **kwargs):
        for s in self.children:
            s.Refresh(*args, **kwargs)
    
    def SetClim(self, *args, **kwargs):
        for s in self.children:
            s.SetClim(*args, **kwargs)
    
    def _GetColormap(self):
        return self.children[0].colormap
    def _SetColormap(self, value):
        for s in self.children:
            s._SetColormap(value)
    
    def _EnableColormap(self, texUnit=0):
        return self.children[0]._EnableColormap(texUnit)
    def _DisableColormap(self):
        return self.children[0]._DisableColormap()
    
    def _GetClim(self):
        return self.children[0].clim
    def _SetClim(self, value):
        for s in self.children:
            s._SetClim(value)
    
    @Property
    def renderStyle():
        """ Not available for SliceTextures. This
        property is implemented to be able to produce a warning when
        it is used.
        """
        def fget(self):
            return 'None'
        def fset(self, value):
            print('Warning: SliceTexture instances have no renderStyle.')
        return locals()
    
    @Property
    def isoThreshold():
        """ Not available for SliceTextures. This
        property is implemented to be able to produce a warning when
        it is used.
        """
        def fget(self):
            return 0.0
        def fset(self, value):
            print('Warning: SliceTexture instances have no isoThreshold.')
        return locals()
    
    @Property
    def interpolate():
        """ Get/Set whether to interpolate the image when zooming in
        (using linear interpolation).
        """
        def fget(self):
            return self.children[0].interpolate
        def fset(self, value):
            for s in self.children:
                s.interpolate = value
        return locals()
    
    @Property
    def index():
        """ The index of the slice in the volume to display.
        """
        def fget(self):
            return self.children[0].index
        def fset(self, value):
            for s in self.children:
                s.index = value
        return locals()
    
    @Property
    def axis():
        """ The axis of the slice in the volume to display.
        """
        def fget(self):
            return self.children[0].axis
        def fset(self, value):
            for s in self.children:
                s.axis = value
        return locals()
    
    @Property
    def edgeColor():
        """ The color of the edge of the slice (can be None).
        """
        def fget(self):
            return self.children[0].edgeColor
        def fset(self, value):
            for s in self.children:
                s.edgeColor = value
        return locals()
    
    @Property
    def edgeColor2():
        """ The color of the edge of the slice when interacting.
        """
        def fget(self):
            return self.children[0].edgeColor2
        def fset(self, value):
            for s in self.children:
                s.edgeColor2 = value
        return locals()



# -*- coding: utf-8 -*-
# Copyright (C) 2012, Almar Klein
#
# Visvis is distributed under the terms of the (new) BSD License.
# The full license can be found in 'license.txt'.

from visvis.utils.pypoints import Pointset, is_Point, is_Pointset
import numpy as np

import visvis as vv


def plot(data1, data2=None, data3=None,
            lw=1, lc='b', ls="-", mw=7, mc='b', ms='', mew=1, mec='k',
            alpha=1, axesAdjust=True, axes=None, **kwargs):
    """ plot(*args, lw=1, lc='b', ls="-", mw=7, mc='b', ms='', mew=1, mec='k',
            alpha=1, axesAdjust=True, axes=None):
    
    Plot 1, 2 or 3 dimensional data and return the Line object.
    
    Usage
    -----
      * plot(Y, ...) plots a 1D signal, with the values plotted along the y-axis
      * plot(X, Y, ...) also supplies x coordinates
      * plot(X, Y, Z, ...) also supplies z coordinates
      * plot(P, ...) plots using a Point or Pointset instance
    
    Keyword arguments
    -----------------
    (The longer names for the line properties can also be used)
    lw : scalar
        lineWidth. The width of the line. If zero, no line is drawn.
    mw : scalar
        markerWidth. The width of the marker. If zero, no marker is drawn.
    mew : scalar
        markerEdgeWidth. The width of the edge of the marker.
    lc : 3-element tuple or char
        lineColor. The color of the line. A tuple should represent the RGB
        values between 0 and 1. If a char is given it must be
        one of 'rgbmcywk', for reg, green, blue, magenta, cyan, yellow,
        white, black, respectively.
    mc : 3-element tuple or char
        markerColor. The color of the marker. See lineColor.
    mec : 3-element tuple or char
        markerEdgeColor. The color of the edge of the marker.
    ls : string
        lineStyle. The style of the line. (See below)
    ms : string
        markerStyle. The style of the marker. (See below)
    axesAdjust : bool
        If axesAdjust==True, this function will call axes.SetLimits(), set
        the camera type to 2D when plotting 2D data and to 3D when plotting
        3D data. If daspectAuto has not been set yet, it is set to True.
    axes : Axes instance
        Display the image in this axes, or the current axes if not given.
    
    Line styles
    -----------
      * Solid line: '-'
      * Dotted line: ':'
      * Dashed line: '--'
      * Dash-dot line: '-.' or '.-'
      * A line that is drawn between each pair of points: '+'
      * No line: '' or None.
    
    Marker styles
    -------------
      * Plus: '+'
      * Cross: 'x'
      * Square: 's'
      * Diamond: 'd'
      * Triangle (pointing up, down, left, right): '^', 'v', '<', '>'
      * Pentagram star: 'p' or '*'
      * Hexgram: 'h'
      * Point/cirle: 'o' or '.'
      * No marker: '' or None
    
    """
    
    # create a dict from the properties and combine with kwargs
    tmp     = { 'lineWidth':lw,         'lineColor':lc,     'lineStyle':ls,
                'markerWidth':mw,       'markerColor':mc,   'markerStyle':ms,
                'markerEdgeWidth':mew,  'markerEdgeColor':mec}
    for i in tmp:
        if not i in kwargs:
            kwargs[i] = tmp[i]
    
    # init dimension variable
    camDim = 0
    
    ##  create the data
    
    # If one argument is given, and it looks like a pointset stored
    # in a numpy array, use it as such
    if isinstance(data1, np.ndarray) and (data2 is None) and (data3 is None):
        if data1.ndim == 2 and data1.shape[1] in (2,3):
            data1 = Pointset(data1)  # Use shape as given
    
    if is_Pointset(data1):
        pp = data1
    elif is_Point(data1):
        pp = Pointset(data1.ndim)
        pp.append(data1)
    else:
        if data1 is None:
            raise Exception("The first argument cannot be None!")
        data1 = np.asanyarray(data1)
        
        d3 = data3
        if data3 is None:
            data3 = 0.1*np.ones(data1.shape)
            camDim = 2
        else:
            camDim = 3
            data3 = np.asanyarray(data3)
        
        if data2 is None:
            if d3 is not None:
                tmp = "third argument in plot() ignored, as second not given."
                print("Warning: " + tmp)
            # y data is given, xdata must be a range starting from 1
            data2 = data1
            data1 = np.arange(1,data2.shape[0]+1)
            data3 = 0.1*np.ones(data2.shape)
        else:
            data2 = np.asanyarray(data2)
        
        # check dimensions
        L = data1.size
        if L != data2.size or L != data3.size:
            raise Exception("Array dimensions do not match! %i vs %i vs %i" %
                    (data1.size, data2.size, data3.size))
        
        # build points
        data1 = data1.reshape((data1.size,1))
        data2 = data2.reshape((data2.size,1))
        data3 = data3.reshape((data3.size,1))
        
        # Concatenate to a single Nx3 array (take care of masked arrays)
        tmp = data1, data2, data3
        if any([isinstance(d, np.ma.MaskedArray) for d in tmp]):
            pp = np.ma.concatenate(tmp, 1)
        else:
            pp = np.concatenate(tmp, 1)
        
    
    # Process camdim for given points or pointsets
    if not camDim:
        camDim = max(2, pp.ndim)
        
    
    ## create the line
    if axes is None:
        axes = vv.gca()
    l = vv.Line(axes, pp)
    l.lw = kwargs['lineWidth']
    l.lc = kwargs['lineColor']
    l.ls = kwargs['lineStyle']
    l.mw = kwargs['markerWidth']
    l.mc = kwargs['markerColor']
    l.ms = kwargs['markerStyle']
    l.mew = kwargs['markerEdgeWidth']
    l.mec = kwargs['markerEdgeColor']
    l.alpha = alpha
    
    ## done...
    if axesAdjust:
        if axes.daspectAuto is None:
            axes.daspectAuto = True
        axes.cameraType = str(camDim)+'d'
        axes.SetLimits()
    
    axes.Draw()
    return l


if __name__ == '__main__':
    vv.figure()
    vv.subplot(311)
    vv.plot([1,3,1,4,1,5,1,6,1,7,6,5,4,3,2,1]) # Passing 1D data
    vv.subplot(312)
    vv.plot([3,4,6,7],[7,5,4,6]) # Passing 2D data
    vv.subplot(313)
    vv.plot([3,4,6,7],[7,5,4,6], [1,2,3,2]) # Passing 3D data



# -*- coding: utf-8 -*-
# Copyright (C) 2012, Almar Klein
#
# Visvis is distributed under the terms of the (new) BSD License.
# The full license can be found in 'license.txt'.

import numpy as np

def statistics(data):
    """ Get StatData object for the given data.
    """
    return StatData(data)

class StatData:
    """ StatData(data)
    
    Get statistics on the given sequence of data. The data must be
    something that is accepted by np.array(). Any nonfinite data points
    (NaN, Inf, -Inf) are removed from the data.
    
    Allows easy calculation of statistics such as mean, std, median,
    any percentile, histogram data and kde's.
    
    This class was initially designed for the boxplot functionality.
    This interface was made public because it can be usefull to the
    generic user too.
    
    """
    
    def __init__(self, data):
        
        # Make numpy array, sort, make 1D, store
        if not isinstance(data, np.ndarray):
            data = np.array(data)
        self._data = np.sort(data.ravel())
        
        # Remove invalids
        valid = np.isfinite(self._data)
        self._data = self._data[valid]
        
        # Calculate some stats
        self._mean = self._data.mean()
        self._std = self._data.std()
    
    def __repr__(self):
        return '<StatData object with %i elements>' % self.size
    
    def __str__(self):
        s = 'Summary of StatData object:\n'
        for key in ['size', 'dmin', 'dmax', 'drange', 'mean', 'std',
                    'Q1', 'Q2', 'Q3', 'IQR']:
            value = str(getattr(self, key))
            line = key.rjust(10) + ' = ' + value + '\n'
            s += line
        return s
    
    @property
    def size(self):
        """ Get the number of elements in the data.
        """
        return self._data.size
    
    @property
    def drange(self):
        """ Get the range of the data (max-min).
        """
        return float(self._data[-1] - self._data[0])
    
    @property
    def dmin(self):
        """ Get the minimum of the data.
        """
        return self._data[0]
    
    @property
    def dmax(self):
        """ Get the max of the data.
        """
        return self._data[-1]
    
    @property
    def mean(self):
        """ Get the mean of the data.
        """
        return self._mean
    
    @property
    def std(self):
        """ Get the standard deviation of the data.
        """
        return self._std
    
    @property
    def Q1(self):
        """ Get first quartile of the data (i.e. the 25th percentile).
        """
        return self.percentile(0.25)
    
    @property
    def Q2(self):
        """ Get second quartile of the data (i.e. the 50th percentile).
        This is the median.
        """
        return self.percentile(0.50)
    
    @property
    def median(self):
        """ Get the median. This is the same as Q2.
        """
        return self.percentile(0.50)
    
    @property
    def Q3(self):
        """ Get second quartile of the data (i.e. the 50th percentile).
        """
        return self.percentile(0.75)
    
    @property
    def IQR(self):
        """ Get the inter-quartile range; the range where 50% of the
        data is.
        """
        return self.percentile(0.75) - self.percentile(0.25)
    
    
    def percentile(self, per, interpolate=True):
        """ percentile(per, interpolate=True)
        
        Given a percentage (as a number between 0 and 1)
        return the value corresponding to that percentile.
        
        By default, the value is linearly interpolated if it does not
        fall exactly on an existing value.
        """
        
        # Calculate float index
        data = self._data
        i = (data.size-1) * float(per)
        
        # Sample the value
        if interpolate:
            # http://en.wikipedia.org/wiki/Percentile
            ik = int(i)
            ir = i - ik
            if ik >= data.size-1:
                return data[ik]
            else:
                return data[ik] + ir*(data[ik+1] - data[ik])
        else:
            i = int(round(i))
            return data[i]
    
    
    def histogram_np(self, nbins=None, drange=None, normed=False, weights=None):
        """"  histogram_np(self, nbins=None, range=None, normed=False, weights=None)
        
        Calculate the histogram of the data.
        
        If nbins is not given, a good value is chosen using
        the Freedman-Diaconis rule.
        Returns a 2-element tuple containing the bin centers and the
        counts.
        
        See also the kde() method.
        
        Parameters
        ----------
        nbins : int or sequence of scalars, optional
            If `bins` is an int, it defines the number of equal-width bins in
            the given range. If `bins` is a sequence, it defines
            the bin edges, including the rightmost edge, allowing for non-uniform
            bin widths. If not given, the optimal number of bins is calculated
            using the Freedman-Diaconis rule.
        range : (float, float)
            The lower and upper range of the bins. If not provided, range is
            simply (a.min(), a.max()). Values outside the range are ignored.
        normed : bool
            If False, the result will contain the number of samples in each bin.
            If True, the result is the value of the probability *density*
            function at the bin, normalized such that the *integral* over the
            range is 1. Note that the sum of the histogram values will not be
            equal to 1 unless bins of unity width are chosen; it is not a
            probability *mass* function.
        weights : array_like
            An array of weights, of the same shape as `a`. Each value in `a`
            only contributes its associated weight towards the bin count
            (instead of 1). If `normed` is True, the weights are normalized,
            so that the integral of the density over the range remains 1.
        
        """
        if nbins is None:
            nbins = self.best_number_of_bins()
        
        # Get histogram
        values, edges = np.histogram(self._data, nbins, drange, normed, weights)
        
        # The bins are the left bin edges, let's get the centers
        centers = np.empty(values.size, np.float32)
        for i in range(len(values)):
            centers[i] = (edges[i] + edges[i+1]) * 0.5
        
        # Done
        return centers, values
    
    
    def histogram(self, nbins=None):
        """  histogram(self, nbins=None)
        
        Calculate the (normalized) histogram of the data.
        
        If nbins is not given, a good value is chosen using
        the Freedman-Diaconis rule.
        
        See also the histogram_np() and kde() methods.
        
        """
        if nbins is None:
            nbins = self.best_number_of_bins()
        
        return self._kernel_density_estimate(nbins, [1])
    
    
    def kde(self, nbins=None, kernel=None):
        """  kde(self, nbins=None)
        
        Calculate the kernel density estimation of the data.
        
        If nbins is not given, a good value is chosen using
        the Freedman-Diaconis rule. If kernel is not given,
        a Gaussian kernel is used, with a sigma depending
        on nbins.
        
        """
        best_nbins = self.best_number_of_bins()
        
        if nbins is None:
            nbins = 4 * best_nbins
        if kernel is None:
            kernel = float(nbins) / best_nbins
        
        return self._kernel_density_estimate(nbins, kernel)
    
    
    def best_number_of_bins(self, minbins=8, maxbins=256):
        """ best_number_of_bins(minbins=8, maxbins=256)
        
        Calculates the best number of bins to make a histogram
        of this data, according to Freedman-Diaconis rule.
        
        """
        # Get data
        data = self._data
        
        # Get number of bins according to Freedman-Diaconis rule
        bin_size = 2 * self.IQR * data.size**(-1.0/3)
        nbins = self.drange / (bin_size+0.001)
        nbins = max(minbins, min(maxbins, int(nbins)))
        
        # Done
        return nbins
    
    
    def _kernel_density_estimate(self, n, kernel):
        """ kernel density estimate(n, kernel)
        """
        
        # Get data
        data = self._data
        
        # Get some statistics
        dmin, dmax, drange = self.dmin, self.dmax, self.drange
        if not drange:
            # A single value, or all values the same
            eps = 0.5
            dmin, dmax, drange = dmin-eps, dmax+eps, 2*eps
        
        # Construct kernel
        if isinstance(kernel, float):
            sigma = kernel
            ktail = int(sigma*3)
            kn = ktail*2 + 1
            t = np.arange(-kn/2.0+0.5, kn/2.0, 1.0, dtype=np.float64)
            k = np.exp(- t**2 / (2*sigma**2) )
        else:
            k = np.array(kernel, dtype='float64').ravel()
            ktail = int( k.size/2 )
        
        # nbins = 3
        #  ___  ___  ___
        # |   ||   ||   |
        # | | || | || | |
        #
        # [--- drange --]
        # [---] dbin
        
        # Get data "bins", these are the bin centers
        dbin = drange / n
        nbins = n + ktail * 2
        bins = (np.arange(nbins) - ktail + 0.5) * dbin + dmin
        
        # Normalize kernel
        k /= k.sum() * data.size * dbin
        
        # Splat the kernels in!
        # kde represents the counts for each bin.
        # xxi represents the data, but scaled to the bin indices. The elements
        # in xxi are sorted (because data is), and if data.size >> nbins
        # there are many equal values in a row. We make use of that to make
        # the bin splatting very efficient.
        kde = np.zeros_like(bins)
        xx = (data-dmin) * (1.0/dbin) # no +ktail here, no j-ktail at kernel index
        xxi = (xx).astype('int32')
        #
        # Init binary search
        step = max(1, int(data.size/n))
        i0, i1, i2 = 0, 0, step
        val = xxi[i0]
        totalSplats = 0
        #
        i2 = min(i2, xxi.size-1)
        while i1<i2:
            if xxi[i2] > val:
                if xxi[i2-1] == val:
                    # === found it!
                    # Splat kernel
                    nSplats = i2-i0
                    kde[val:val+k.size] += k * nSplats
                    totalSplats += nSplats
                    # Reset for next value
                    step = max(1, nSplats)
                    i0, i1, i2 = i2, i2, i2+step
                    val = xxi[i0]
                else:
                    # Too far
                    step = max(1, int(0.5*step))
                    i2 = i1 + step
            else:
                # Not far enough
                i1, i2 = i2, i2+step
            # Not beyond end!
            i2 = min(i2, xxi.size-1)
        # Wrap up
        if True:
            # Correct kernel (it might be at the way right edge
            if val >= n:
                val -= 1
            # Splat kernel
            nSplats = i2-i0+1
            kde[val:val+k.size] += k * nSplats
            totalSplats += nSplats
        
        # Check (should be equal)
        #print(totalSplats, xxi.size)
        
        if False:
            # Analog algoritm, better readable perhaps, but much slower:
            for x in data:
                i = (x-dmin) * (1.0/dbin) + ktail
                i = int(round(i))
                for j in range(k.size):
                    kde[i+j-ktail] += k[j]
        
        # Done
        return bins, kde



# -*- coding: utf-8 -*-
# Copyright (C) 2012, Almar Klein
#
# Visvis is distributed under the terms of the (new) BSD License.
# The full license can be found in 'license.txt'.

import visvis as vv
from visvis.utils.pypoints import Pointset


def solidBox(translation=None, scaling=None, direction=None, rotation=None,
                axesAdjust=True, axes=None):
    """ solidBox(translation=None, scaling=None, direction=None, rotation=None,
                    axesAdjust=True, axes=None)
    
    Creates a solid cube (or box if you scale it) centered at the
    origin. Returns an OrientableMesh.
    
    Parameters
    ----------
    Note that translation, scaling, and direction can also be given
    using a Point instance.
    translation : (dx, dy, dz), optional
        The translation in world units of the created world object.
    scaling: (sx, sy, sz), optional
        The scaling in world units of the created world object.
    direction: (nx, ny, nz), optional
        Normal vector that indicates the direction of the created world object.
    rotation: scalar, optional
        The anle (in degrees) to rotate the created world object around its
        direction vector.
    axesAdjust : bool
        If True, this function will call axes.SetLimits(), and set
        the camera type to 3D. If daspectAuto has not been set yet,
        it is set to False.
    axes : Axes instance
        Display the bars in the given axes, or the current axes if not given.
    
    """
    
    # Create vertices of a cube
    pp = Pointset(3)
    # Bottom
    pp.append(-0.5,-0.5,-0.5)
    pp.append(+0.5,-0.5,-0.5)
    pp.append(+0.5,+0.5,-0.5)
    pp.append(-0.5,+0.5,-0.5)
    # Top
    pp.append(-0.5,-0.5,+0.5)
    pp.append(-0.5,+0.5,+0.5)
    pp.append(+0.5,+0.5,+0.5)
    pp.append(+0.5,-0.5,+0.5)
    
    # Init vertices and normals
    vertices = Pointset(3)
    normals = Pointset(3)
    
    # Create vertices
    for i in [3,2,1,0]: # Top
        vertices.append(pp[i]); normals.append(0,0,-1)
    for i in [7,6,5,4]: # Bottom
        vertices.append(pp[i]); normals.append(0,0,+1)
    for i in [5,6,2,3]: # Front
        vertices.append(pp[i]); normals.append(0,+1,0)
    for i in [1,7,4,0]: # Back
        vertices.append(pp[i]); normals.append(0,-1,0)
    for i in [4,5,3,0]: # Left
        vertices.append(pp[i]); normals.append(-1,0,0)
    for i in [2,6,7,1]: # Right
        vertices.append(pp[i]); normals.append(+1,0,0)
    
    
    ## Visualize
    
    # Create axes
    if axes is None:
        axes = vv.gca()
    
    # Create mesh and set orientation
    m = vv.OrientableMesh(axes, vertices, None, normals, verticesPerFace=4)
    #
    if translation is not None:
        m.translation = translation
    if scaling is not None:
        m.scaling = scaling
    if direction is not None:
        m.direction = direction
    if rotation is not None:
        m.rotation = rotation
    
    # Adjust axes
    if axesAdjust:
        if axes.daspectAuto is None:
            axes.daspectAuto = False
        axes.cameraType = '3d'
        axes.SetLimits()
    
    # Done
    axes.Draw()
    return m


if __name__ == '__main__':
    m1 = vv.solidBox((3,1,1), (2,2,1), rotation=-20)
    m2 = vv.solidBox((1,1,0), (1,1,1.5), direction=(1,0.4,0.2))
    m1.faceColor = 'r'
    m2.faceColor = 'g'



# -*- coding: utf-8 -*-
# Copyright (C) 2012, Almar Klein
#
# Visvis is distributed under the terms of the (new) BSD License.
# The full license can be found in 'license.txt'.

import visvis as vv
from visvis.utils.pypoints import Pointset, is_Pointset
from visvis.processing import lineToMesh

def solidLine(pp, radius=1.0, N=16, axesAdjust=True, axes=None):
    """ solidLine(pp, radius=1.0, N=16, axesAdjust=True, axes=None)
    
    Creates a solid line in 3D space.
    
    Parameters
    ----------
    Note that translation, scaling, and direction can also be given
    using a Point instance.
    pp : Pointset
        The sequence of points of which the line consists.
    radius : scalar or sequence
        The radius of the line to create. If a sequence if given, it
        specifies the radius for each point in pp.
    N : int
        The number of subdivisions around its centerline. If smaller
        than 8, flat shading is used instead of smooth shading.
    axesAdjust : bool
        If True, this function will call axes.SetLimits(), and set
        the camera type to 3D. If daspectAuto has not been set yet,
        it is set to False.
    axes : Axes instance
        Display the bars in the given axes, or the current axes if not given.
    
    """
    
    # Check first argument
    if is_Pointset(pp):
        pass
    else:
        raise ValueError('solidLine() needs a Pointset or list of pointsets.')
    
    # Obtain mesh and make a visualization mesh
    baseMesh = lineToMesh(pp, radius, N)
    
    
    ## Visualize
    
    # Get axes
    if axes is None:
        axes = vv.gca()
    
    # Create mesh object
    m = vv.Mesh(axes, baseMesh)
    
    # Adjust axes
    if axesAdjust:
        if axes.daspectAuto is None:
            axes.daspectAuto = False
        axes.cameraType = '3d'
        axes.SetLimits()
    
    # Return
    axes.Draw()
    return m
    
    
    
if __name__ == '__main__':
    # Create series of points
    pp = Pointset(3)
    pp.append(0,1,0)
    pp.append(3,2,1)
    pp.append(4,5,2)
    pp.append(2,3,1)
    pp.append(0,4,0)
    #pp.append(0,1,0) # Circular
    # Make a surface-line with varying diameter
    m = vv.solidLine(pp, [0.1, 0.2, 0.3, 0.1, 0.2], 8)



# -*- coding: utf-8 -*-
# Copyright (C) 2012, Almar Klein
#
# SSDF is distributed under the terms of the (new) BSD License.
# See http://www.opensource.org/licenses/bsd-license.php

""" ssdf.ssdf_bin.py

Implements functionality to read/write binary ssdf (.bsdf) files.
"""

import struct
import zlib

from . import ClassManager
from .ssdf_base import Struct, VirtualArray, SSDFReader, SSDFWriter, Block, _CLASS_NAME
from .ssdf_base import np, binary_type, ascii_type


# Formatters for struct (un)packing
_SMALL_NUMBER_FMT = '<B'
_LARGE_NUMBER_FMT = '<Q'
_TYPE_FMT = '<B'
_PARTITION_LEN_FMT = '<I'
_PARTITION_SIZE = 2**20 # 1 MB

# Types for binary
_TYPE_NONE = ord('N')
_TYPE_INT = ord('I')
_TYPE_FLOAT = ord('F')
_TYPE_UNICODE = ord('U')
_TYPE_ARRAY = ord('A')
_TYPE_LIST = ord('L')
_TYPE_DICT = ord('D')


class BinarySSDFReader(SSDFReader):
    
   
    def read_binary_blocks(self, f):
        """ read_binary_blocks(f)
        
        Given a file, creates the block instances.
        This is a generator function.
        
        """
        count = 0
        while True:
            count += 1
            
            # Get type. If no bytes left, we're done
            try:
                type_id, = struct.unpack(_TYPE_FMT, f.read(1))
            except StopIteration:
                break
            
            # Get indentation
            indent = f.read_number()
            
            # Get name
            name_len = f.read_number()
            if name_len:
                name = f.read(name_len).decode('utf-8')
            else:
                name = None
            
            # Get data
            data_len = f.read_number()
            data = f.read(data_len)
            
            # Create block instance
            yield BinaryBlock(indent, count, name, type_id, data=data)
    
    
    def read(self, file_or_bytes):
        """ read(file_or_bytes)
        
        Given a file or bytes, convert it to a struct by reading the
        blocks, building the tree and converting each block to its
        Python object.
        
        """
        
        # Get file
        if isinstance(file_or_bytes, binary_type):
            f = _VirtualFile(file_or_bytes)
        else:
            f = file_or_bytes
        
        # Check header
        bb1 = "BSDF".encode('utf-8')
        try:
            bb2 = f.read(len(bb1))
        except Exception:
            raise ValueError('Could not read header of binary SSDF file.')
        if bb1 != bb2:
            raise ValueError('Given SSDF bytes/file does not have the right header.')
        
        # Create compressed file to read from
        fc = _CompressedFile(f)
        
        # Create blocks and build tree
        root = BinaryBlock(-1, -1, type=_TYPE_DICT)
        block_gen = self.read_binary_blocks(fc)
        self.build_tree(root, block_gen)
        
        # Convert to real objects and return
        return root.to_object()
    

class BinarySSDFWriter(SSDFWriter):
    
    
    def write_binary_blocks(self, blocks, f):
        """ write_binary_blocks(blocks, f)
        
        Writes the given blocks to a binary file.
        
        """
        
        # Skip first, it is the root object
        for block in blocks[1:]:
            
            # Write type.
            f.write(struct.pack(_TYPE_FMT, block._type))
            
            # Write indentation
            f.write_number(block._indent)
            
            # Write name
            if block._name:
                name = block._name.encode('utf-8')
                f.write_number(len(name))
                f.write(name)
            else:
                f.write_number(0)
            
            # Write data
            if isinstance(block._data, list):
                data_len = sum([len(d) for d in block._data])
                f.write_number(data_len)
                for part in block._data:
                    f.write(part)
            else:
                data_len = len(block._data)
                f.write_number(data_len)
                f.write(block._data)
    
    
    
    def write(self, object, f=None):
        """ write(object, f=None)
        
        Serializes the given struct. If a file is given, writes bytes
        to that file, otherwise returns a bytes instance.
        
        """
        
        # Check input
        if f is None:
            f = _VirtualFile()
            return_bytes = True
        else:
            return_bytes = False
        
        # Write header
        f.write('BSDF'.encode('utf-8'))
        
        # Make compressed file
        fc = _CompressedFile(f)
        
        # Create block object
        root = BinaryBlock.from_object(-1, binary_type(), object)
        
        # Collect blocks and write
        blocks = self.flatten_tree(root)
        self.write_binary_blocks(blocks, fc)
        fc.flush()
        
        # Return?
        if return_bytes:
            return f.get_bytes()
    


class BinaryBlock(Block):
    
    
    def to_object(self):
        # Determine what type of object we are dealing with using the
        # type id.
        type = self._type
        if type==_TYPE_INT: return self._to_int()
        elif type==_TYPE_FLOAT: return self._to_float()
        elif type==_TYPE_UNICODE: return self._to_unicode()
        elif type==_TYPE_ARRAY: return self._to_array()
        elif type==_TYPE_LIST: return self._to_list()
        elif type==_TYPE_DICT: return self._to_dict()
        elif type==_TYPE_NONE: return self._to_none()
        else:
            print("SSDF: invalid type %s in block %i." % (repr(type), self._blocknr))
            return None
    
    
    def _from_none(self, value=None):
        self._type = _TYPE_NONE
        self._data = binary_type()
    
    def _to_none(self):
        return None
    
    
    def _from_int(self, value):
        self._type = _TYPE_INT
        self._data = struct.pack('<q', int(value))
    
    def _to_int(self):
        return struct.unpack('<q', self._data)[0]
    
    
    def _from_float(self, value):
        self._type = _TYPE_FLOAT
        self._data = struct.pack('<d', float(value))
    
    def _to_float(self):
        return struct.unpack('<d', self._data)[0]
    
    
    def _from_unicode(self, value):
        self._type = _TYPE_UNICODE
        self._data = value.encode('utf-8')
    
    def _to_unicode(self):
        return self._data.decode('utf-8')
    
    
    def _from_array(self, value):
        self._type = _TYPE_ARRAY
        f =_VirtualFile()
        # Write shape and dtype
        f.write_number(value.ndim)
        for s in value.shape:
            f.write_number(s)
        f.write_string(str(value.dtype))
        # Write data
        # tostring() returns bytes, not a string on py3k
        self._data = [f.get_bytes(), value.tostring()]
    
    def _to_array(self):
        f = _VirtualFile(self._data)
        # Get shape and dtype
        ndim = f.read_number()
        shape = [f.read_number() for i in range(ndim)]
        dtypestr = ascii_type(f.read_string())
        # Create numpy array or Virtual array
        i = f._fp
        if not np:
            return VirtualArray(shape, dtypestr, self._data[i:])
        else:
            if i < len(self._data):
                value = np.frombuffer(self._data, dtype=dtypestr, offset=i)
            else:
                value = np.array([], dtype=dtypestr)
            if np.prod(shape) == value.size:
                value.shape = tuple(shape)
            else:
                print("SSDF: prod(shape)!=size on line %i."%self._blocknr)
            return value
    
    
    def _from_dict(self, value):
        self._type = _TYPE_DICT
        self._data = binary_type()
        # Get keys sorted by name
        keys = [key for key in value]
        keys.sort()
        # Process children
        for key in keys:
            # Skip all the buildin stuff
            if key.startswith("__"):
                continue
            # We have the key, go get the value!
            val = value[key]
            # Skip methods, or anything else we can call
            if hasattr(val,'__call__') and not hasattr(val, '__to_ssdf__'):
                # Note: py3.x does not have function callable
                continue
            # Add!
            subBlock = BinaryBlock.from_object(self._indent+1, key, val)
            self._children.append(subBlock)
    
    def _to_dict(self):
        value = Struct()
        for child in self._children:
            val = child.to_object()
            if child._name:
                value[child._name] = val
            else:
                print("SSDF: unnamed element in dict in block %i."%child._blocknr)
        # Make class instance?
        if _CLASS_NAME in value:
            className = value[_CLASS_NAME]
            if className in ClassManager._registered_classes:
                value = ClassManager._registered_classes[className].__from_ssdf__(value)
            else:
                print("SSDF: class %s not registered." % className)
        # Done
        return value
    
    
    
    def _from_list(self, value):
        self._type = _TYPE_LIST
        self._data = binary_type()
        
        # Process children
        for element in value:
            # Add element as subblock
            subBlock = BinaryBlock.from_object(self._indent+1, None, element)
            self._children.append(subBlock)
    
    def _to_list(self):
        value = []
    
        for child in self._children:
            val = child.to_object()
            if child._name:
                print("SSDF: named element in list in block %i."%child._blocknr)
            else:
                value.append(val)
        return value



class _FileWithExtraMethods:
    
    def write_number(self, n):
        if n < 255:
            self.write( struct.pack(_SMALL_NUMBER_FMT, n) )
        else:
            self.write( struct.pack(_SMALL_NUMBER_FMT, 255) )
            self.write( struct.pack(_LARGE_NUMBER_FMT, n) )
    
    def write_bytes(self, bb):
        self.write_number(len(bb))
        self.write(bb)
    
    def write_string(self, ss):
        self.write_bytes(ss.encode('utf-8'))
    
    def read_number(self):
        n, = struct.unpack(_SMALL_NUMBER_FMT, self.read(1))
        if n == 255:
            n, = struct.unpack(_LARGE_NUMBER_FMT, self.read(8))
        return n
    
    def read_bytes(self):
        n = self.read_number()
        return self.read(n)
    
    def read_string(self):
        return self.read_bytes().decode('utf-8')


class _VirtualFile(_FileWithExtraMethods):
    """ _VirtualFile(bb=None)
    
    Wraps a bytes instance to provide a file-like interface. Also
    represents a file like object to which bytes can be written, and
    the resulting bytes can be obtained using get_bytes().
    
    """
    def __init__(self, bb=None):
        # For reading
        self._bb = bb
        self._fp = 0
        # For writing
        self._parts = []
    
    def read(self, n):
        i1 = self._fp
        self._fp = i2 = self._fp + n
        return self._bb[i1:i2]
    
    def write(self, data):
        self._parts.append(data)
    
    def close(self):
        pass
    
    def get_bytes(self):
        return binary_type().join(self._parts)


class _CompressedFile(_FileWithExtraMethods):
    """ _CompressedFile(file)
    
    Wraps a file object to transparantly support reading and writing
    data from/to a compressed file.
    
    Data is compressed in partitions of say 1MB. A partition in the file
    consists of a small header and a body. The header consists of 4 bytes
    representing the body's length (little endian unsigned 32 bit int).
    The body consists of bytes compressed using DEFLATE (i.e. zip).
    
    """
    
    def __init__(self, f):
        
        # Store file
        self._file = f
        
        # For reading
        self._buffer = binary_type()
        self._bp = 0 # buffer pointer
        
        # For writing
        self._parts = []
        self._pp = 0 # parts pointer (is more like a counter)
    
    
    def _read_new_partition(self):
        """ _read_new_partition()
        
        Returns the data in the next partition.
        Returns false if no new partition is available.
        
        """
        
        # Get bytes and read partition length
        # If eof, return True
        bb = self._file.read(4)
        if not bb:
            self._buffer = binary_type()
            self._bp = 0
            return False
        n, = struct.unpack(_PARTITION_LEN_FMT, bb)
        
        # Read partition and decompress
        bb = self._file.read(n)
        data = zlib.decompress(bb)
        del bb
        
        # Done
        return data
    
    
    def _write_new_partition(self):
        """ _write_new_partition()
        
        Compress the buffered data and write to file. Reset buffer.
        
        """
        
        # Get data
        data = binary_type().join(self._parts)
        
        # Reset buffer
        self._parts = []
        self._pp = 0
        
        # Compress and write
        bb = zlib.compress(data)
        del data
        header = struct.pack(_PARTITION_LEN_FMT, len(bb))
        self._file.write(header)
        self._file.write(bb)
    
    
    def read(self, n):
        """ read(n)
        
        Read n bytes. Partitions are automatically decompressed on the fly.
        If the end of the file is reached, raises StopIteration.
        
        """
        
        # Get bytes in buffer
        bytes_in_buffer = len(self._buffer) - self._bp
        
        if bytes_in_buffer < n:
            # Read partitions untill we have enough
            
            # Prepare
            localBuffer = [self._buffer[self._bp:]]
            bufferCount = len(localBuffer[0])
            
            # Collect blocks of data
            while bufferCount < n:
                partition = self._read_new_partition()
                if partition:
                    localBuffer.append(partition)
                    bufferCount += len(partition)
                else:
                    raise StopIteration
            
            # Update buffer
            offset = len(partition) - (bufferCount - n)
            self._bp = offset
            self._buffer = partition
            
            # Truncate last block and get data
            localBuffer[-1] = partition[:offset]
            data = binary_type().join(localBuffer)
        
        else:
            # Get data from current partition
            i1 = self._bp
            self._bp = i2 = i1 + n
            data = self._buffer[i1:i2]
        
        return data
    
    
    def write(self, data):
        """ write(data)
        
        Write data. The data is buffered until the accumulated size
        exceeds the partition size. When this happens, the data is compressed
        and written to the real underlying file/data.
        
        """
        
        # Get how many bytes we have beyond _PARTITION_SIZE
        i = (self._pp + len(data)) - _PARTITION_SIZE
        
        if i > 0:
            # Add portion to buffer, store remainder
            self._parts.append(data[:i])
            data = data[i:]
            # Write the buffer away and reset buffer
            self._write_new_partition()
        
        # Add data to buffer
        self._parts.append(data)
        self._pp += len(data)
    
    
    def flush(self):
        """ flush()
        
        After the last write, use this to compress and write
        the last partition.
        
        """
        self._write_new_partition()



# -*- coding: utf-8 -*-
# Copyright (C) 2011, Nicolas P. Rougier
# Copyright (C) 2012, Robert Schroll
# Copyright (C) 2012, Almar Klein
#
# This module contains code taken from freetype-py, which was modified
# to integrate with visvis.

""" Text rendering using the FreeType font renderer.

A note about availability of fonts:
  * We make sure that there is always a "sans" font by shipping it with visvis.
    This also makes for consistent looks.
  * We try our best to also provide a "serif" and "mono" font.
  * Any other font requires fc-match (which is usually available on Linux)

"""

import os
import sys
import math
import numpy as np
import OpenGL.GL as gl

from visvis.core.misc import getResourceDir, getOpenGlCapable

import subprocess

from visvis.core.shaders import Shader, ShaderCodePart

from visvis.text.text_base import AtlasTexture, FontManager
from visvis.text.text_base import correctVertices, simpleTextureDraw

from visvis.text.freetype import (Face, Vector, Matrix, FT_KERNING_UNFITTED,
                                  FT_LOAD_RENDER, FT_LOAD_FORCE_AUTOHINT )

try:
    n2_16 = long(0x10000)  # noqa
except Exception:
    n2_16 = 0x10000

# The scale factor for textures. The texture glyphs are made TEX_SCALE times
# bigger than the screen resolution. This means that we prevent very blurry
# pieces for small text. The text becomes a bit too crisp I think, but
# I suspect that when we apply the full screen aliasing, the text will look
# great!
TEX_SCALE = 2.5

# todo: have pyzo ship freeType lib on Windows and use that if possible.
# todo: When we implement full screen antialiasing, we can remove the shader here

FRAGMENT_SHADER_ = """
// Uniforms obtained from OpenGL
    uniform sampler2D texture; // The 3D texture
    uniform vec2 shape; // And its shape (as in OpenGl)
    
    void main()
    {
        // Get centre location
        vec2 pos = gl_TexCoord[0].xy;
        
        // Init value
        vec4 color1 = vec4(0.0, 0.0, 0.0, 0.0);
        
        // Init kernel and number of steps
        vec4 kernel = vec4 (0.2, 0.2, 0.1 , 0.1);
        //vec4 kernel = vec4 (0.3, 0.2, 0.1 , 0.0);
        int sze = 3;
        
        // Init step size in tex coords
        float dx = 1.0/shape.x;
        float dy = 1.0/shape.y;
        
        // Convolve
        for (int y=-sze; y<sze+1; y++)
        {
            for (int x=-sze; x<sze+1; x++)
            {
                float k = kernel[int(abs(float(x)))] * kernel[int(abs(float(y)))];
                vec2 dpos = vec2(float(x)*dx, float(y)*dy);
                color1 += texture2D(texture, pos+dpos) * k;
            }
        }
        gl_FragColor = color1 * gl_Color;
        
    }
"""
FRAGMENT_SHADER = """
// Uniforms obtained from OpenGL
    uniform sampler2D texture; // The 3D texture
    uniform vec2 shape; // And its shape (as in OpenGl)
    
    void main()
    {
        // Get centre location
        vec2 pos = gl_TexCoord[0].xy;
        
        // Define kernel. Chose such that k0+2*k1 == 1
        // k0 >> k1 is sharp   k0 ~ k1 is blurry
        // Optimal values depend on TEX_SCALE too!
        float k0 = 0.5;
        float k1 = 0.25;
        
        // Init step size in tex coords
        float dx = 1.0/shape.x;
        float dy = 1.0/shape.y;
        
        vec4 color1 = vec4(0.0, 0.0, 0.0, 0.0);
        
        color1 += texture2D(texture, pos+vec2(-dx,-dy) ) * k1 * k1;
        color1 += texture2D(texture, pos+vec2(-dx,0.0) ) * k1 * k0;
        color1 += texture2D(texture, pos+vec2(-dx,+dy) ) * k1 * k1;
        
        color1 += texture2D(texture, pos+vec2(0.0,-dy) ) * k0 * k1;
        color1 += texture2D(texture, pos+vec2(0.0,0.0) ) * k0 * k0;
        color1 += texture2D(texture, pos+vec2(0.0,+dy) ) * k0 * k1;
        
        color1 += texture2D(texture, pos+vec2(+dx,-dy) ) * k1 * k1;
        color1 += texture2D(texture, pos+vec2(+dx,0.0) ) * k1 * k0;
        color1 += texture2D(texture, pos+vec2(+dx,+dy) ) * k1 * k1;
        
        // Set final color
        gl_FragColor = color1 * gl_Color;
        
    }
"""

class FreeTypeAtlas(AtlasTexture):
    '''
    Taken from freetype-py's TextureAtlas
    Group multiple small data regions into a larger texture.

    The algorithm is based on the article by Jukka Jylänki : "A Thousand Ways
    to Pack the Bin - A Practical Approach to Two-Dimensional Rectangle Bin
    Packing", February 27, 2010. More precisely, this is an implementation of
    the Skyline Bottom-Left algorithm based on C++ sources provided by Jukka
    Jylänki at: http://clb.demon.fi/files/RectangleBinPack/

    Example usage:
    --------------

    atlas = TextureAtlas(512,512,3)
    region = atlas.get_region(20,20)
    ...
    atlas.set_region(region, data)
    '''

    def __init__(self, width=1024, height=1024, depth=1):
        '''
        Initialize a new atlas of given size.

        Parameters
        ----------

        width : int
            Width of the underlying texture

        height : int
            Height of the underlying texture

        depth : 1 or 3
            Depth of the underlying texture
        '''
        AtlasTexture.__init__(self)
        
        self.width  = int(math.pow(2, int(math.log(width, 2) + 0.5)))
        self.height = int(math.pow(2, int(math.log(height, 2) + 0.5)))
        self.depth  = depth
        self.nodes  = [ (0,0,self.width), ]
        self.data   = np.zeros((self.height, self.width, self.depth),
                               dtype=np.uint8)
        
        self.used   = 0
    
    
    def upload(self):
        '''
        Upload atlas data into video memory.
        '''
        # Note that we only uplad one channel
        self.SetData(self.data[:,:,0])

    

    def set_region(self, region, data):
        '''
        Set a given region width provided data.

        Parameters
        ----------

        region : (int,int,int,int)
            an allocated region (x,y,width,height)

        data : numpy array
            data to be copied into given region
        '''

        x, y, width, height = region
        self.data[y:y+height,x:x+width, :] = data



    def get_region(self, width, height):
        '''
        Get a free region of given size and allocate it

        Parameters
        ----------

        width : int
            Width of region to allocate

        height : int
            Height of region to allocate

        Return
        ------
            A newly allocated region as (x,y,width,height) or (-1,-1,0,0)
        '''
        width, height = int(width), int(height)  # issue #89
        best_height = self.data.shape[0] * 10
        best_index = -1
        best_width = self.data.shape[1] * 10
        region = 0, 0, width, height

        for i in range(len(self.nodes)):
            y = self.fit(i, width, height)
            if y >= 0:
                node = self.nodes[i]
                if (y+height < best_height or (y+height == best_height and node[2] < best_width)):
                    best_height = y+height
                    best_index = i
                    best_width = node[2]
                    region = node[0], y, width, height

        if best_index == -1:
            return -1,-1,0,0
        
        node = region[0], region[1]+height, width
        self.nodes.insert(best_index, node)

        i = best_index+1
        while i < len(self.nodes):
            node = self.nodes[i]
            prev_node = self.nodes[i-1]
            if node[0] < prev_node[0]+prev_node[2]:
                shrink = prev_node[0]+prev_node[2] - node[0]
                x,y,w = self.nodes[i]
                self.nodes[i] = x+shrink, y, w-shrink
                if self.nodes[i][2] <= 0:
                    del self.nodes[i]
                    i -= 1
                else:
                    break
            else:
                break
            i += 1

        self.merge()
        self.used += width*height
        return region



    def fit(self, index, width, height):
        '''
        Test if region (width,height) fit into self.nodes[index]

        Parameters
        ----------

        index : int
            Index of the internal node to be tested

        width : int
            Width or the region to be tested

        height : int
            Height or the region to be tested

        '''

        node = self.nodes[index]
        x,y = node[0], node[1]
        width_left = width
        
        if x+width > self.width:
            return -1

        i = index
        while width_left > 0:
            node = self.nodes[i]
            y = max(y, node[1])
            if y+height > self.height:
                return -1
            width_left -= node[2]
            i += 1
        return y



    def merge(self):
        '''
        Merge nodes
        '''

        i = 0
        while i < len(self.nodes)-1:
            node = self.nodes[i]
            next_node = self.nodes[i+1]
            if node[1] == next_node[1]:
                self.nodes[i] = node[0], node[1], node[2]+next_node[2]
                del self.nodes[i+1]
            else:
                i += 1



class FreeTypeFontManager(FontManager):
    
    def __init__(self):
        FontManager.__init__(self)
        
        self._font_names = {}
        self._fonts = {}
        self.atlas = FreeTypeAtlas(1024, 1024, 1)
        
        # Create shader
        self._shader = None
    
    @property
    def shader(self):
        if self._shader is None:
            if not getOpenGlCapable('2.0', 'Antialiased text'):
                self._shader = self
                self.Enable = lambda: None
                self.Disable = lambda: None
                global TEX_SCALE
                TEX_SCALE = 1.5 # Make text more blurry
            else:
                # Create shader
                self._shader = Shader()
                # Set fragment code, vertex code is empty
                self._shader.vertex.Clear()
                fragment = ShaderCodePart('textaa', '', FRAGMENT_SHADER)
                self._shader.fragment.AddPart(fragment)
                # Set uniform
                shape = self.atlas.data.shape[:2]
                uniform_shape = [float(s) for s in reversed(list(shape))]
                self.shader.SetStaticUniform('shape', uniform_shape)
        
        return self._shader
    
    def GetFont(self, fontname, size, bold=False, italic=False):
        fontfile = self.get_font_file(fontname, bold, italic)
        sig = fontfile, size
        if sig not in self._fonts:
            self._fonts[sig] = TextureFont(self.atlas, fontfile, size)
        return self._fonts[sig]
    
    def get_font_file(self, fontname, bold, italic):
        
        sig = (fontname, bold, italic)
        if not sig in self._font_names:
            
            # Did we ship this font with visvis?
            if True:
                fname = self.get_font_file_in_resources(fontname, bold, italic)
            
            # Try getting it in a smarter way (platform dependent)
            if not fname:
                if sys.platform.startswith('win'):
                    fname = self.get_font_file_with_windows(fontname, bold, italic)
                else:
                    fname = self.get_font_file_with_fcmatch(fontname, bold, italic)
            
            # Check. If not known, use sans
            if not fname:
                print('Warning: cannot retrieve font file fole for %s.' % fontname)
                fname = self.get_font_file('sans', bold, italic)
            
            # Store
            self._font_names[sig] = fname
        
        # Done
        return self._font_names[sig]
    
    def get_font_file_in_resources(self, fontname, bold, italic):
        
        # Normalize name and attributes
        fontname = fontname[0].upper() + fontname[1:].lower()
        bold = 'Bold' if bold else ''
        italic = 'Oblique' if italic else ''
        
        # Be a bit smart
        if fontname in ['Xkcd', 'Humor']:
            fname = 'HumorSans.otf'  # xkcd-style font
        else:
            fname = 'Free' + fontname + bold + italic + '.otf'  # Freetype font
        
        # Check if exist
        fname = os.path.join( getResourceDir(), fname )
        if os.path.exists(fname):
            return fname
        else:
            return ''
    
    def get_font_file_with_fcmatch(self, fontname, bold, italic):
        # fc-match is (almost?) always present in Linux.
        # fc-match is available on Mac, not sure if installed by default
        
        weight = 200 if bold else 80
        slant = 100 if italic else 0
        try:
            args = ['fc-match', '-f', '%{file}',
                        '%s:weight=%i:slant=%i' % (fontname, weight, slant)]
            fname, err = subprocess.Popen(args, stdout=subprocess.PIPE).communicate()
            #py3k only: fname = subprocess.check_output(args)
            return fname.decode('utf-8') # Return as string
        except OSError:
            return ''
    
    def get_font_file_with_windows(self, fontname, bold, italic):
        
        # On Windows we know some fonts
        M = {'sans': 'arial', 'serif': 'times', 'mono': 'cour'}
        
        # Prepare
        postfix = ''
        if bold and italic: postfix = 'bi'
        elif bold: postfix = 'bd' # What is the d for?
        elif italic: postfix = 'i'
        
        # Select
        fname = M.get(fontname.lower(), '')
        if fname:
            fname = 'C:/Windows/Fonts/' + fname + postfix + '.ttf'
        else:
            # Just use the given name
            fname = 'C:/Windows/Fonts/' + fontname + '.ttf'
        
        # Check and return
        if not os.path.isfile(fname):
            fname = ''
        return fname
    
    def Compile(self, textObject):
        
        # Get text string with escaped text converted to Unicode
        tt = self.ConvertEscapedText(textObject.text)
        
        # Init arrays
        vertices = np.zeros((len(tt)*4,3), dtype=np.float32)
        texcoords = np.zeros((len(tt)*4,2), dtype=np.float32)
        
        # Prepare
        textObject._shift = None
        pen = [0,0]
        prev = None
        
        # Calculate font size
        # note: I can also imagine doing it the other way around; i.e. textSize
        # becomes as FreeType sees it, and we scale the fonts in the
        # prerendered text renderer. We'd have to change all the
        # uses fontSize though. So this is simply easier.
        fontSize = textObject.fontSize * 1.4 * TEX_SCALE
        fig = textObject.GetFigure()
        if fig:
            fontSize *= fig._relativeFontSize
        
        # Store tex_scale, integer fontsize and residu
        textObject._tex_scale = TEX_SCALE
        textObject._actualFontSize = int(round(fontSize))
        
        fonts = [(self.GetFont(textObject.fontName, textObject._actualFontSize), 0, False, False)]
        font, voffset, bold, italic = fonts[-1]
        escaped = False
        for i,charcode in enumerate(tt):
            if not escaped:
                if charcode == '_':
                    font = self.GetFont(textObject.fontName, font.size*0.7, bold, italic)
                    voffset -= 0.1*font.size
                    prev = None  # Disable kerning
                    continue
                elif charcode == '^':
                    font = self.GetFont(textObject.fontName, font.size*0.7, bold, italic)
                    voffset += 0.5*font.size
                    prev = None
                    continue
                elif charcode == '\x06':
                    italic = True
                    font = self.GetFont(textObject.fontName, font.size, bold, italic)
                    continue
                elif charcode == '\x07':
                    bold = True
                    font = self.GetFont(textObject.fontName, font.size, bold, italic)
                    continue
                elif charcode == '{':
                    fonts.append((font, voffset, bold, italic))
                    continue
                elif charcode == '}':
                    if len(fonts) > 1:
                        fonts.pop()
                        font, voffset, bold, italic = fonts[-1]
                        continue
                elif charcode == '\\':
                    if i < len(tt)-1 and tt[i+1] in '_^{}':
                        escaped = True
                        continue
            glyph = font[charcode]
            if glyph is None:
                continue # Character not available
            kerning = glyph.get_kerning(prev)
            x0 = pen[0] + glyph.offset[0] + kerning
            y0 = pen[1] + glyph.offset[1] + voffset
            x1 = x0 + glyph.size[0]
            y1 = y0 - glyph.size[1]
            u0 = glyph.texcoords[0]
            v0 = glyph.texcoords[1]
            u1 = glyph.texcoords[2]
            v1 = glyph.texcoords[3]

            # index     = i*4
            _vertices  = [[x0,y0,1],[x0,y1,1],[x1,y1,1], [x1,y0,1]]
            _texcoords = [[u0,v0],[u0,v1],[u1,v1], [u1,v0]]

            vertices[i*4:i*4+4] = _vertices
            texcoords[i*4:i*4+4] = _texcoords
            pen[0] = pen[0]+glyph.advance[0]/64.0 + kerning
            pen[1] = pen[1]+glyph.advance[1]/64.0
            prev = charcode
            font, voffset, bold, italic = fonts[-1]
            escaped = False
        
        # Flip and shift vertices
        vertices /= TEX_SCALE
        vertices *= (1, -1, 1)
        vertices += (0, font.ascender/TEX_SCALE, 0)
        
        # Store width
        if False:  # glyph is not None: # Not used
            textObject.width = pen[0]-glyph.advance[0]/64.0+glyph.size[0] if tt else 0
        
        # Update dynamic texture
        self.atlas.upload()
        
        # Store data.
        textObject._SetCompiledData(vertices, texcoords)
    
    
    def Position(self, textObject):
        FontManager.Position(self, textObject)
        
        # Get data
        vertices, texcoords = textObject._GetCompiledData()
        vertices = vertices.copy()
        
        # Use default algorithm to correct the vertices for alginment and angle
        font = self.GetFont(textObject.fontName, textObject._actualFontSize)
        correctVertices(textObject, vertices, font.height/TEX_SCALE)
        
        # Store
        textObject._SetFinalData(vertices, texcoords)
    
    
    def Draw(self, textObject, x=0, y=0, z=0):
        # Check if we need to recompile in case TEX_SCALE was changed
        if hasattr(textObject, '_tex_scale') and textObject._tex_scale != TEX_SCALE:
            textObject.Invalidate()
        
        FontManager.Draw(self, textObject)
        
        # Get data
        vertices, texCords = textObject._GetFinalData()
        
        # Translate
        if x or y or z:
            gl.glPushMatrix()
            gl.glTranslatef(x, y, z)
        
        # Draw
        self.shader.Enable()
        simpleTextureDraw(vertices, texCords, self.atlas, textObject.textColor)
        self.shader.Disable()
        
        # Un-translate
        if x or y or z:
            gl.glPopMatrix()
    

class TextureFont:
    '''
    A texture font gathers a set of glyph relatively to a given font filename
    and size.
    '''

    def __init__(self, atlas, filename, size):
        '''
        Initialize font

        Parameters:
        -----------

        atlas: TextureAtlas
            Texture atlas where glyph texture will be stored

        filename: str
            Font filename
        
        size : float
            Font size
        '''
        self.atlas = atlas
        self.filename = filename.encode('utf-8') # Make bytes
        self.size = size
        self.glyphs = {}
        face = Face( self.filename )
        face.set_char_size( int(self.size*64))
        self._dirty = False
        metrics = face.size
        self.ascender  = metrics.ascender/64.0
        self.descender = metrics.descender/64.0
        self.height    = metrics.height/64.0
        self.linegap   = self.height - self.ascender + self.descender
        self.depth = atlas.depth
    

    def __getitem__(self, charcode):
        '''
        x.__getitem__(y) <==> x[y]
        '''
        if charcode not in self.glyphs.keys():
            self.load('%c' % charcode)
        return self.glyphs.get(charcode, None)

 
    def load(self, charcodes = ''):
        '''
        Build glyphs corresponding to individual characters in charcodes.

        Parameters:
        -----------
        
        charcodes: [str | unicode]
            Set of characters to be represented
        '''
        face = Face( self.filename )
        try:
            return self._load(charcodes, face)
        finally:
            # Make sure to clear the face. I've seen some pretty bad
            # crashes on Windows when an exception was thrown in the code
            # in _load(), because the hold all the variables, preventing the
            # face from being cleared.
            face.__del__()
            face._FT_Face = None
    
    def _load(self, charcodes, face):
        pen = Vector(0,0)
        hres = 16*72
        hscale = 1.0/16
        # todo: use set_pixel_sizes?
        for charcode in charcodes:
            face.set_char_size( int(self.size * 64), 0, hres, 72 )
            matrix = Matrix( int((hscale) * n2_16), int((0.0) * n2_16),
                             int((0.0)    * n2_16), int((1.0) * n2_16) )
            face.set_transform( matrix, pen )
            if charcode in self.glyphs.keys():
                continue

            self.dirty = True
            flags = FT_LOAD_RENDER | FT_LOAD_FORCE_AUTOHINT

            face.load_char( charcode, flags )
            bitmap = face.glyph.bitmap
            left   = face.glyph.bitmap_left
            top    = face.glyph.bitmap_top
            width  = face.glyph.bitmap.width
            rows   = face.glyph.bitmap.rows
            pitch  = face.glyph.bitmap.pitch
            padding = 1 # Extra space for neighboring pixels to join in aa
            margin = 1 # Extra space to prevent overlap from other glyphs
            padmarg = padding + margin
            
            x,y,w,h = self.atlas.get_region(width/self.depth+2*padmarg,
                                                        rows+2*padmarg)
            if x < 0:
                print('Missed !')
                continue
            x,y = x+padmarg, y+padmarg
            w,h = w-2*padmarg, h-2*padmarg
            data = []
            for i in range(rows):
                data.extend(bitmap.buffer[i*pitch:i*pitch+width])
            data = np.array(data,dtype=np.ubyte).reshape(h,w,self.depth)
            gamma = 1.5
            Z = ((data/255.0)**(gamma))
            data = (Z*255).astype(np.ubyte)
            if True: # Add an extra pixel, otherwise there's no room for the aa
                # Note that we can do this because we asked for a larger region anyway
                data2 = np.zeros((  data.shape[0]+2*padding,
                                    data.shape[1]+2*padding,
                                    data.shape[2]), np.ubyte)
                data2[padding:-padding,padding:-padding,:] = data
                data = data2
                x,y = x-padding, y-padding
                w,h = w+2*padding, h+2*padding
            
            self.atlas.set_region((x,y,w,h), data)

            # Build glyph
            size   = w,h
            offset = left, top
            advance= face.glyph.advance.x, face.glyph.advance.y

            u0     = (x +     0.0)/float(self.atlas.width)
            v0     = (y +     0.0)/float(self.atlas.height)
            u1     = (x + w - 0.0)/float(self.atlas.width)
            v1     = (y + h - 0.0)/float(self.atlas.height)
            texcoords = (u0,v0,u1,v1)
            glyph = TextureGlyph(charcode, size, offset, advance, texcoords)
            self.glyphs[charcode] = glyph

            # Generate kerning
            for g in self.glyphs.values():
                # 64 * 64 because of 26.6 encoding AND the transform matrix used
                # in texture_font_load_face (hres = 64)
                kerning = face.get_kerning(g.charcode, charcode, mode=FT_KERNING_UNFITTED)
                if kerning.x != 0:
                    glyph.kerning[g.charcode] = kerning.x/(64.0*64.0)
                kerning = face.get_kerning(charcode, g.charcode, mode=FT_KERNING_UNFITTED)
                if kerning.x != 0:
                    g.kerning[charcode] = kerning.x/(64.0*64.0)

            # High resolution advance.x calculation
            # gindex = face.get_char_index( charcode )
            # a = face.get_advance(gindex, FT_LOAD_RENDER | FT_LOAD_TARGET_LCD)/(64*72)
            # glyph.advance = a, glyph.advance[1]


class TextureGlyph:
    '''
    A texture glyph gathers information relative to the size/offset/advance and
    texture coordinates of a single character. It is generally built
    automatically by a TextureFont.
    '''

    def __init__(self, charcode, size, offset, advance, texcoords):
        '''
        Build a new texture glyph

        Parameter:
        ----------

        charcode : char
            Represented character

        size: tuple of 2 ints
            Glyph size in pixels

        offset: tuple of 2 floats
            Glyph offset relatively to anchor point

        advance: tuple of 2 floats
            Glyph advance

        texcoords: tuple of 4 floats
            Texture coordinates of bottom-left and top-right corner
        '''
        self.charcode = charcode
        self.size = size
        self.offset = offset
        self.advance = advance
        self.texcoords = texcoords
        self.kerning = {}


    def get_kerning(self, charcode):
        ''' Get kerning information

        Parameters:
        -----------

        charcode: char
            Character preceding this glyph
        '''
        if charcode in self.kerning.keys():
            return self.kerning[charcode]
        else:
            return 0



#!/usr/bin/env python

import visvis as vv
app = vv.use()

def ensureString(s):
    if isinstance(s, str):
        return s
    else:
        return s.decode('ascii')

# get info
version, vendor, renderer, ext = vv.getOpenGlInfo()

# print!
print('Information about the OpenGl version on this system:\n')
print('version:    ' + version)
print('vendor:     ' + vendor)
print('renderer:   ' + renderer)
print('extensions: ' + str(len(ext.split())) + ' different extensions')

# Wait
try:
    input('\nPress enter to continue...') # Python 3
except NameError:
    raw_input('\nPress enter to continue...') # noqa - Python 2



# -*- coding: utf-8 -*-
# Copyright (C) 2012, Almar Klein
#
# Visvis is distributed under the terms of the (new) BSD License.
# The full license can be found in 'license.txt'.

import visvis as vv
import os

# Try importing imageio
imageio = None

try:
    import imageio
except ImportError:
    pass


def imread(filename):
    """ imread(filename)
    
    Read image from file or http, requires imageio.
    
    """
    
    if imageio is None:
        raise RuntimeError("visvis.imread requires the imageio package.")
    
    if not os.path.isfile(filename) and '//' not in filename:
        # try loading from the resource dir
        path = vv.misc.getResourceDir()
        filename2 = os.path.join(path, filename)
        if os.path.isfile(filename2):
            filename = filename2
        else:
            pass  # imageio can read from http and more ...
    
    return imageio.imread(filename)


if __name__ == '__main__':
    im = vv.imread('astronaut.png')
    t = vv.imshow(im)



#!/usr/bin/env python

""" Example that demonstrates how movies can be recorded
and exported to gif/swf/avi.

This is not an interactive example, but a script.

"""

import visvis as vv

# Create something to show, let's show a red teapot!
mesh = vv.solidTeapot()
mesh.faceColor = 'r'

# Prepare
Nangles = 36
a = vv.gca()
f = vv.gcf()
rec = vv.record(a)

# Rotate camera
for i in range(Nangles):
    a.camera.azimuth = 360 * float(i) / Nangles
    if a.camera.azimuth>180:
        a.camera.azimuth -= 360
    a.Draw() # Tell the axes to redraw
    f.DrawNow() # Draw the figure NOW, instead of waiting for GUI event loop


# Export
rec.Stop()
rec.Export('teapot.gif')


""" NOTES

A note on a.Draw()
------------------

This is necessary to tell the axes to redraw itself instead of using its
buffered image. Changing a property of the axes or any object inside it
will automatically trigger a redraw, but this is not so for the camera,
since it's a bit of a low level thing...

More explicit (and better documented) control over the camera may be
implemented later.


A note on f.DrawNow()
---------------------

This is necessary to invoke a redraw at THAT moment. Normal draw calls
(such as the one to a.Draw(), will post a paint request to the GUI toolkit,
which is handled when the GUI event loop returns. In this example, that would
be never, since we do not start an event loop :)

"""



#!/usr/bin/env python
""" This example illustrates visualizing a four dimensional data set
(a series of 3D volumes). The same method can be used to visualize
the motion of lines, images, meshes, etc.
"""
import visvis as vv

# create multiple instances of the same volume (simulate motion)
vols = [vv.aVolume()]
for i in range(9):
    vol = vols[i].copy()
    vol[2:] = vol[:-2]
    vol[:2] = 0
    vols.append(vol)

# create figure, axes, and data container object
f = vv.clf()
a = vv.gca()
m = vv.MotionDataContainer(a)

# create volumes, loading them into opengl memory, and insert into container.
for vol in vols:
    t = vv.volshow(vol)
    t.parent = m
    t.colormap = vv.CM_HOT
    # Remove comments to use iso-surface rendering
    #t.renderStyle = 'iso'
    #t.isoThreshold = 0.2

# set some settings
a.daspect = 1,1,-1
a.xLabel = 'x'
a.yLabel = 'y'
a.zLabel = 'z'

# Enter main loop
app = vv.use()
app.Run()



# -*- coding: utf-8 -*-
#   Copyright (C) 2012, Almar Klein
#
#   This code is subject to the (new) BSD license:
#
#   Redistribution and use in source and binary forms, with or without
#   modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the <organization> nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL <COPYRIGHT HOLDER> BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

""" Module images2swf

Provides a function (writeSwf) to store a series of PIL images or numpy
arrays in an SWF movie, that can be played on a wide range of OS's.

This module came into being because I wanted to store a series of images
in a movie that can be viewed by other people, and which I can embed in
flash presentations. For writing AVI or MPEG you really need a c/c++
library, and allthough the filesize is then very small, the quality is
sometimes not adequate. Besides I'd like to be independant of yet another
package. I tried writing animated gif using PIL (which is widely available),
but the quality is so poor because it only allows for 256 different colors.
[EDIT: thanks to Ant1, now the quality of animated gif isn't so bad!]
I also looked into MNG and APNG, two standards similar to the PNG stanard.
Both standards promise exactly what I need. However, hardly any application
can read those formats, and I cannot import them in flash.

Therefore I decided to check out the swf file format, which is very well
documented. This is the result: a pure python module to create an SWF file
that shows a series of images. The images are stored using the DEFLATE
algorithm (same as PNG and ZIP and which is included in the standard Python
distribution). As this compression algorithm is much more effective than
that used in GIF images, we obtain better quality (24 bit colors + alpha
channel) while still producesing smaller files (a test showed ~75%).
Although SWF also allows for JPEG compression, doing so would probably
require a third party library (because encoding JPEG is much harder).

This module requires Python 2.x and numpy.

sources and tools:

- SWF on wikipedia
- Adobes "SWF File Format Specification" version 10
  (http://www.adobe.com/devnet/swf/pdf/swf_file_format_spec_v10.pdf)
- swftools (swfdump in specific) for debugging
- iwisoft swf2avi can be used to convert swf to avi/mpg/flv with really
  good quality, while file size is reduced with factors 20-100.
  A good program in my opinion. The free version has the limitation
  of a watermark in the upper left corner.

"""

import os, sys, time
import zlib

try:
    import numpy as np
except ImportError:
    np = None

try:
    import PIL.Image
except ImportError:
    PIL = None


# True if we are running on Python 3.
# Code taken from six.py by Benjamin Peterson (MIT licensed)
import types
PY3 = sys.version_info[0] == 3
if PY3:
    string_types = str,
    integer_types = int,
    class_types = type,
    text_type = str
    binary_type = bytes
else:
    string_types = basestring,
    integer_types = (int, long)
    class_types = (type, types.ClassType)
    text_type = unicode
    binary_type = str


# todo: use imageio/FreeImage to support reading JPEG images from SWF?


def checkImages(images):
    """ checkImages(images)
    Check numpy images and correct intensity range etc.
    The same for all movie formats.
    """
    # Init results
    images2 = []

    for im in images:
        if PIL and isinstance(im, PIL.Image.Image):
            # We assume PIL images are allright
            images2.append(im)

        elif np and isinstance(im, np.ndarray):
            # Check and convert dtype
            if im.dtype == np.uint8:
                images2.append(im) # Ok
            elif im.dtype in [np.float32, np.float64]:
                theMax = im.max()
                if theMax > 128 and theMax < 300:
                    pass # assume 0:255
                else:
                    im = im.copy()
                    im[im < 0] = 0
                    im[im > 1] = 1
                    im *= 255
                images2.append(im.astype(np.uint8))
            else:
                im = im.astype(np.uint8)
                images2.append(im)
            # Check size
            if im.ndim == 2:
                pass # ok
            elif im.ndim == 3:
                if im.shape[2] not in [3, 4]:
                    raise ValueError('This array can not represent an image.')
            else:
                raise ValueError('This array can not represent an image.')
        else:
            raise ValueError('Invalid image type: ' + str(type(im)))

    # Done
    return images2


## Base functions and classes


class BitArray:
    """ Dynamic array of bits that automatically resizes
    with factors of two.
    Append bits using .Append() or +=
    You can reverse bits using .Reverse()
    """

    def __init__(self, initvalue=None):
        self.data = np.zeros((16,), dtype=np.uint8)
        self._len = 0
        if initvalue is not None:
            self.Append(initvalue)

    def __len__(self):
        return self._len  # self.data.shape[0]

    def __repr__(self):
        return self.data[:self._len].tostring().decode('ascii')

    def _checkSize(self):
        # check length... grow if necessary
        arraylen = self.data.shape[0]
        if self._len >= arraylen:
            tmp = np.zeros((arraylen*2,), dtype=np.uint8)
            tmp[:self._len] = self.data[:self._len]
            self.data = tmp

    def __add__(self, value):
        self.Append(value)
        return self

    def Append(self, bits):

        # check input
        if isinstance(bits, BitArray):
            bits = str(bits)
        if isinstance(bits, int):
            bits = str(bits)
        if not isinstance(bits, string_types):
            raise ValueError("Append bits as strings or integers!")

        # add bits
        for bit in bits:
            self.data[self._len] = ord(bit)
            self._len += 1
            self._checkSize()

    def Reverse(self):
        """ In-place reverse. """
        tmp = self.data[:self._len].copy()
        self.data[:self._len] = np.flipud(tmp)

    def ToBytes(self):
        """ Convert to bytes. If necessary,
        zeros are padded to the end (right side).
        """
        bits = str(self)

        # determine number of bytes
        nbytes = 0
        while nbytes * 8 < len(bits):
            nbytes += 1
        # pad
        bits = bits.ljust(nbytes * 8, '0')

        # go from bits to bytes
        bb = binary_type()
        for i in range(nbytes):
            tmp = int(bits[i * 8: (i + 1) * 8], 2)
            bb += intToUint8(tmp)

        # done
        return bb


if PY3:
    def intToUint32(i):
        return int(i).to_bytes(4, 'little')

    def intToUint16(i):
        return int(i).to_bytes(2, 'little')

    def intToUint8(i):
        return int(i).to_bytes(1, 'little')
else:
    def intToUint32(i):
        number = int(i)
        n1, n2, n3, n4 = 1, 256, 256 * 256, 256 * 256 * 256
        b4, number = number // n4, number % n4
        b3, number = number // n3, number % n3
        b2, number = number // n2, number % n2
        b1 = number
        return chr(b1) + chr(b2) + chr(b3) + chr(b4)

    def intToUint16(i):
        i = int(i)
        # devide in two parts (bytes)
        i1 = i % 256
        i2 = int(i // 256)
        # make string (little endian)
        return chr(i1) + chr(i2)

    def intToUint8(i):
        return chr(int(i))


def intToBits(i, n=None):
    """ convert int to a string of bits (0's and 1's in a string),
    pad to n elements. Convert back using int(ss,2). """
    ii = i

    # make bits
    bb = BitArray()
    while ii > 0:
        bb += str(ii % 2)
        ii = ii >> 1
    bb.Reverse()

    # justify
    if n is not None:
        if len(bb) > n:
            raise ValueError("intToBits fail: len larger than padlength.")
        bb = str(bb).rjust(n, '0')

    # done
    return BitArray(bb)


def bitsToInt(bb, n=8):
    # Init
    value = ''

    # Get value in bits
    for i in range(len(bb)):
        b = bb[i:i+1]
        tmp = bin(ord(b))[2:]
        #value += tmp.rjust(8,'0')
        value = tmp.rjust(8, '0') + value

    # Make decimal
    return(int(value[:n], 2))


def getTypeAndLen(bb):
    """ bb should be 6 bytes at least
    Return (type, length, length_of_full_tag)
    """
    # Init
    value = ''

    # Get first 16 bits
    for i in range(2):
        b = bb[i:i + 1]
        tmp = bin(ord(b))[2:]
        #value += tmp.rjust(8,'0')
        value = tmp.rjust(8, '0') + value

    # Get type and length
    type = int(value[:10], 2)
    L = int(value[10:], 2)
    L2 = L + 2

    # Long tag header?
    if L == 63: # '111111'
        value = ''
        for i in range(2, 6):
            b = bb[i:i + 1]  # becomes a single-byte bytes() on both PY3 and PY2
            tmp = bin(ord(b))[2:]
            #value += tmp.rjust(8,'0')
            value = tmp.rjust(8, '0') + value
        L = int(value, 2)
        L2 = L + 6

    # Done
    return type, L, L2


def signedIntToBits(i, n=None):
    """ convert signed int to a string of bits (0's and 1's in a string),
    pad to n elements. Negative numbers are stored in 2's complement bit
    patterns, thus positive numbers always start with a 0.
    """

    # negative number?
    ii = i
    if i < 0:
        # A negative number, -n, is represented as the bitwise opposite of
        ii = abs(ii) - 1  # the positive-zero number n-1.

    # make bits
    bb = BitArray()
    while ii > 0:
        bb += str(ii % 2)
        ii = ii >> 1
    bb.Reverse()

    # justify
    bb = '0' + str(bb) # always need the sign bit in front
    if n is not None:
        if len(bb) > n:
            raise ValueError("signedIntToBits fail: len larger than padlength.")
        bb = bb.rjust(n, '0')

    # was it negative? (then opposite bits)
    if i < 0:
        bb = bb.replace('0', 'x').replace('1', '0').replace('x', '1')

    # done
    return BitArray(bb)


def twitsToBits(arr):
    """ Given a few (signed) numbers, store them
    as compactly as possible in the wat specifief by the swf format.
    The numbers are multiplied by 20, assuming they
    are twits.
    Can be used to make the RECT record.
    """

    # first determine length using non justified bit strings
    maxlen = 1
    for i in arr:
        tmp = len(signedIntToBits(i*20))
        if tmp > maxlen:
            maxlen = tmp

    # build array
    bits = intToBits(maxlen, 5)
    for i in arr:
        bits += signedIntToBits(i * 20, maxlen)

    return bits


def floatsToBits(arr):
    """ Given a few (signed) numbers, convert them to bits,
    stored as FB (float bit values). We always use 16.16.
    Negative numbers are not (yet) possible, because I don't
    know how the're implemented (ambiguity).
    """
    bits = intToBits(31, 5) # 32 does not fit in 5 bits!
    for i in arr:
        if i < 0:
            raise ValueError("Dit not implement negative floats!")
        i1 = int(i)
        i2 = i - i1
        bits += intToBits(i1, 15)
        bits += intToBits(i2 * 2 ** 16, 16)
    return bits


def _readFrom(fp, n):
    bb = binary_type()
    try:
        while len(bb) < n:
            tmp = fp.read(n-len(bb))
            bb += tmp
            if not tmp:
                break
    except EOFError:
        pass
    return bb


## Base Tag

class Tag:

    def __init__(self):
        self.bytes = binary_type()
        self.tagtype = -1

    def ProcessTag(self):
        """ Implement this to create the tag. """
        raise NotImplemented()

    def GetTag(self):
        """ Calls processTag and attaches the header. """
        self.ProcessTag()

        # tag to binary
        bits = intToBits(self.tagtype, 10)

        # complete header uint16 thing
        bits += '1' * 6  # = 63 = 0x3f
        # make uint16
        bb = intToUint16(int(str(bits), 2))

        # now add 32bit length descriptor
        bb += intToUint32(len(self.bytes))

        # done, attach and return
        bb += self.bytes
        return bb

    def MakeRectRecord(self, xmin, xmax, ymin, ymax):
        """ Simply uses makeCompactArray to produce
        a RECT Record. """
        return twitsToBits([xmin, xmax, ymin, ymax])

    def MakeMatrixRecord(self, scale_xy=None, rot_xy=None, trans_xy=None):

        # empty matrix?
        if scale_xy is None and rot_xy is None and trans_xy is None:
            return "0"*8

        # init
        bits = BitArray()

        # scale
        if scale_xy:
            bits += '1'
            bits += floatsToBits([scale_xy[0], scale_xy[1]])
        else:
            bits += '0'

        # rotation
        if rot_xy:
            bits += '1'
            bits += floatsToBits([rot_xy[0], rot_xy[1]])
        else:
            bits += '0'

        # translation (no flag here)
        if trans_xy:
            bits += twitsToBits([trans_xy[0], trans_xy[1]])
        else:
            bits += twitsToBits([0, 0])

        # done
        return bits


## Control tags

class ControlTag(Tag):
    def __init__(self):
        Tag.__init__(self)


class FileAttributesTag(ControlTag):
    def __init__(self):
        ControlTag.__init__(self)
        self.tagtype = 69

    def ProcessTag(self):
        self.bytes = '\x00'.encode('ascii') * (1+3)


class ShowFrameTag(ControlTag):
    def __init__(self):
        ControlTag.__init__(self)
        self.tagtype = 1

    def ProcessTag(self):
        self.bytes = binary_type()


class SetBackgroundTag(ControlTag):
    """ Set the color in 0-255, or 0-1 (if floats given). """
    def __init__(self, *rgb):
        self.tagtype = 9
        if len(rgb) == 1:
            rgb = rgb[0]
        self.rgb = rgb

    def ProcessTag(self):
        bb = binary_type()
        for i in range(3):
            clr = self.rgb[i]
            if isinstance(clr, float):
                clr = clr * 255
            bb += intToUint8(clr)
        self.bytes = bb


class DoActionTag(Tag):
    def __init__(self, action='stop'):
        Tag.__init__(self)
        self.tagtype = 12
        self.actions = [action]

    def Append(self, action):
        self.actions.append(action)

    def ProcessTag(self):
        bb = binary_type()

        for action in self.actions:
            action = action.lower()
            if action == 'stop':
                bb += '\x07'.encode('ascii')
            elif action == 'play':
                bb += '\x06'.encode('ascii')
            else:
                print("warning, unkown action: %s" % action)

        bb += intToUint8(0)
        self.bytes = bb


## Definition tags
class DefinitionTag(Tag):
    counter = 0  # to give automatically id's

    def __init__(self):
        Tag.__init__(self)
        DefinitionTag.counter += 1
        self.id = DefinitionTag.counter  # id in dictionary


class BitmapTag(DefinitionTag):

    def __init__(self, im):
        DefinitionTag.__init__(self)
        self.tagtype = 36 # DefineBitsLossless2

        # convert image (note that format is ARGB)
        # even a grayscale image is stored in ARGB, nevertheless,
        # the fabilous deflate compression will make it that not much
        # more data is required for storing (25% or so, and less than 10%
        # when storing RGB as ARGB).

        if len(im.shape) == 3:
            if im.shape[2] in [3, 4]:
                tmp = np.ones((im.shape[0], im.shape[1], 4),
                              dtype=np.uint8) * 255
                for i in range(3):
                    tmp[:, :, i + 1] = im[:, :, i]
                if im.shape[2] == 4:
                    tmp[:, :, 0] = im[:, :, 3]  # swap channel where alpha is in
            else:
                raise ValueError("Invalid shape to be an image.")

        elif len(im.shape) == 2:
            tmp = np.ones((im.shape[0], im.shape[1], 4), dtype=np.uint8)*255
            for i in range(3):
                tmp[:, :, i + 1] = im[:, :]
        else:
            raise ValueError("Invalid shape to be an image.")

        # we changed the image to uint8 4 channels.
        # now compress!
        self._data = zlib.compress(tmp.tostring(), zlib.DEFLATED)
        self.imshape = im.shape

    def ProcessTag(self):

        # build tag
        bb = binary_type()
        bb += intToUint16(self.id)   # CharacterID
        bb += intToUint8(5)     # BitmapFormat
        bb += intToUint16(self.imshape[1])   # BitmapWidth
        bb += intToUint16(self.imshape[0])   # BitmapHeight
        bb += self._data            # ZlibBitmapData

        self.bytes = bb


class PlaceObjectTag(ControlTag):
    def __init__(self, depth, idToPlace=None, xy=(0, 0), move=False):
        ControlTag.__init__(self)
        self.tagtype = 26
        self.depth = depth
        self.idToPlace = idToPlace
        self.xy = xy
        self.move = move

    def ProcessTag(self):
        # retrieve stuff
        depth = self.depth
        xy = self.xy
        id = self.idToPlace

        # build PlaceObject2
        bb = binary_type()
        if self.move:
            bb += '\x07'.encode('ascii')
        else:
            bb += '\x06'.encode('ascii')  # (8 bit flags): 4:matrix, 2:character, 1:move
        bb += intToUint16(depth) # Depth
        bb += intToUint16(id) # character id
        bb += self.MakeMatrixRecord(trans_xy=xy).ToBytes() # MATRIX record
        self.bytes = bb


class ShapeTag(DefinitionTag):
    def __init__(self, bitmapId, xy, wh):
        DefinitionTag.__init__(self)
        self.tagtype = 2
        self.bitmapId = bitmapId
        self.xy = xy
        self.wh = wh

    def ProcessTag(self):
        """ Returns a defineshape tag. with a bitmap fill """

        bb = binary_type()
        bb += intToUint16(self.id)
        xy, wh = self.xy, self.wh
        tmp = self.MakeRectRecord(xy[0], wh[0], xy[1], wh[1])  # ShapeBounds
        bb += tmp.ToBytes()

        # make SHAPEWITHSTYLE structure

        # first entry: FILLSTYLEARRAY with in it a single fill style
        bb += intToUint8(1)  # FillStyleCount
        bb += '\x41'.encode('ascii')  # FillStyleType  (0x41 or 0x43, latter is non-smoothed)
        bb += intToUint16(self.bitmapId)  # BitmapId
        #bb += '\x00' # BitmapMatrix (empty matrix with leftover bits filled)
        bb += self.MakeMatrixRecord(scale_xy=(20, 20)).ToBytes()

#         # first entry: FILLSTYLEARRAY with in it a single fill style
#         bb += intToUint8(1)  # FillStyleCount
#         bb += '\x00' # solid fill
#         bb += '\x00\x00\xff' # color

        # second entry: LINESTYLEARRAY with a single line style
        bb += intToUint8(0)  # LineStyleCount
        #bb += intToUint16(0*20) # Width
        #bb += '\x00\xff\x00'  # Color

        # third and fourth entry: NumFillBits and NumLineBits (4 bits each)
        # I each give them four bits, so 16 styles possible.
        bb += '\x44'.encode('ascii')

        self.bytes = bb

        # last entries: SHAPERECORDs ... (individual shape records not aligned)
        # STYLECHANGERECORD
        bits = BitArray()
        bits += self.MakeStyleChangeRecord(0, 1, moveTo=(self.wh[0],
                                                         self.wh[1]))
        # STRAIGHTEDGERECORD 4x
        bits += self.MakeStraightEdgeRecord(-self.wh[0], 0)
        bits += self.MakeStraightEdgeRecord(0, -self.wh[1])
        bits += self.MakeStraightEdgeRecord(self.wh[0], 0)
        bits += self.MakeStraightEdgeRecord(0, self.wh[1])

        # ENDSHAPRECORD
        bits += self.MakeEndShapeRecord()

        self.bytes += bits.ToBytes()

        # done
        #self.bytes = bb

    def MakeStyleChangeRecord(self, lineStyle=None, fillStyle=None,
                              moveTo=None):

        # first 6 flags
        # Note that we use FillStyle1. If we don't flash (at least 8) does not
        # recognize the frames properly when importing to library.

        bits = BitArray()
        bits += '0'  # TypeFlag (not an edge record)
        bits += '0'  # StateNewStyles (only for DefineShape2 and Defineshape3)
        if lineStyle:
            bits += '1'  # StateLineStyle
        else:
            bits += '0'
        if fillStyle:
            bits += '1'  # StateFillStyle1
        else:
            bits += '0'
        bits += '0'  # StateFillStyle0
        if moveTo:
            bits += '1'  # StateMoveTo
        else:
            bits += '0'

        # give information
        # todo: nbits for fillStyle and lineStyle is hard coded.

        if moveTo:
            bits += twitsToBits([moveTo[0], moveTo[1]])
        if fillStyle:
            bits += intToBits(fillStyle, 4)
        if lineStyle:
            bits += intToBits(lineStyle, 4)

        return bits
        #return bitsToBytes(bits)

    def MakeStraightEdgeRecord(self, *dxdy):
        if len(dxdy) == 1:
            dxdy = dxdy[0]

        # determine required number of bits
        xbits = signedIntToBits(dxdy[0] * 20)
        ybits = signedIntToBits(dxdy[1] * 20)
        nbits = max([len(xbits), len(ybits)])

        bits = BitArray()
        bits += '11'  # TypeFlag and StraightFlag
        bits += intToBits(nbits-2, 4)
        bits += '1'  # GeneralLineFlag
        bits += signedIntToBits(dxdy[0] * 20, nbits)
        bits += signedIntToBits(dxdy[1] * 20, nbits)

        # note: I do not make use of vertical/horizontal only lines...

        return bits
        #return bitsToBytes(bits)

    def MakeEndShapeRecord(self):
        bits = BitArray()
        bits += "0"     # TypeFlag: no edge
        bits += "0"*5   # EndOfShape
        return bits
        #return bitsToBytes(bits)


## Last few functions
def buildFile(fp, taglist, nframes=1, framesize=(500, 500), fps=10, version=8):
    """ Give the given file (as bytes) a header. """

    # compose header
    bb = binary_type()
    bb += 'F'.encode('ascii')  # uncompressed
    bb += 'WS'.encode('ascii')  # signature bytes
    bb += intToUint8(version)  # version
    bb += '0000'.encode('ascii')  # FileLength (leave open for now)
    bb += Tag().MakeRectRecord(0, framesize[0], 0, framesize[1]).ToBytes()
    bb += intToUint8(0) + intToUint8(fps)  # FrameRate
    bb += intToUint16(nframes)
    fp.write(bb)

    # produce all tags
    for tag in taglist:
        fp.write(tag.GetTag())

    # finish with end tag
    fp.write('\x00\x00'.encode('ascii'))

    # set size
    sze = fp.tell()
    fp.seek(4)
    fp.write(intToUint32(sze))


def writeSwf(filename, images, duration=0.1, repeat=True):
    """Write an swf-file from the specified images. If repeat is False,
    the movie is finished with a stop action. Duration may also
    be a list with durations for each frame (note that the duration
    for each frame is always an integer amount of the minimum duration.)

    Images should be a list consisting of PIL images or numpy arrays.
    The latter should be between 0 and 255 for integer types, and
    between 0 and 1 for float types.

    """

    # Check Numpy
    if np is None:
        raise RuntimeError("Need Numpy to write an SWF file.")

    # Check images (make all Numpy)
    images2 = []
    images = checkImages(images)
    if not images:
        raise ValueError("Image list is empty!")
    for im in images:
        if PIL and isinstance(im, PIL.Image.Image):
            if im.mode == 'P':
                im = im.convert()
            im = np.asarray(im)
            if len(im.shape) == 0:
                raise MemoryError("Too little memory to convert PIL image to array")
        images2.append(im)

    # Init
    taglist = [FileAttributesTag(), SetBackgroundTag(0, 0, 0)]

    # Check duration
    if hasattr(duration, '__len__'):
        if len(duration) == len(images2):
            duration = [d for d in duration]
        else:
            raise ValueError("len(duration) doesn't match amount of images.")
    else:
        duration = [duration for im in images2]

    # Build delays list
    minDuration = float(min(duration))
    delays = [round(d/minDuration) for d in duration]
    delays = [max(1, int(d)) for d in delays]

    # Get FPS
    fps = 1.0/minDuration

    # Produce series of tags for each image
    t0 = time.time()
    nframes = 0
    for im in images2:
        bm = BitmapTag(im)
        wh = (im.shape[1], im.shape[0])
        sh = ShapeTag(bm.id, (0, 0), wh)
        po = PlaceObjectTag(1, sh.id, move=nframes > 0)
        taglist.extend([bm, sh, po])
        for i in range(delays[nframes]):
            taglist.append(ShowFrameTag())
        nframes += 1

    if not repeat:
        taglist.append(DoActionTag('stop'))

    # Build file
    t1 = time.time()
    fp = open(filename, 'wb')
    try:
        buildFile(fp, taglist, nframes=nframes, framesize=wh, fps=fps)
    except Exception:
        raise
    finally:
        fp.close()
    t2 =  time.time()

    #print("Writing SWF took %1.2f and %1.2f seconds" % (t1-t0, t2-t1) )


def _readPixels(bb, i, tagType, L1):
    """ With pf's seed after the recordheader, reads the pixeldata.
    """

    # Check Numpy
    if np is None:
        raise RuntimeError("Need Numpy to read an SWF file.")

    # Get info
    charId = bb[i:i + 2]; i += 2
    format = ord(bb[i:i + 1]); i += 1
    width = bitsToInt(bb[i:i + 2], 16); i += 2
    height = bitsToInt(bb[i:i + 2], 16); i += 2

    # If we can, get pixeldata and make nunmpy array
    if format != 5:
        print("Can only read 24bit or 32bit RGB(A) lossless images.")
    else:
        # Read byte data
        offset = 2 + 1 + 2 + 2  # all the info bits
        bb2 = bb[i:i+(L1-offset)]

        # Decompress and make numpy array
        data = zlib.decompress(bb2)
        a = np.frombuffer(data, dtype=np.uint8)

        # Set shape
        if tagType == 20:
            # DefineBitsLossless - RGB data
            try:
                a.shape = height, width, 3
            except Exception:
                # Byte align stuff might cause troubles
                print("Cannot read image due to byte alignment")
        if tagType == 36:
            # DefineBitsLossless2 - ARGB data
            a.shape = height, width, 4
            # Swap alpha channel to make RGBA
            b = a
            a = np.zeros_like(a)
            a[:, :, 0] = b[:, :, 1]
            a[:, :, 1] = b[:, :, 2]
            a[:, :, 2] = b[:, :, 3]
            a[:, :, 3] = b[:, :, 0]

        return a


def readSwf(filename, asNumpy=True):
    """Read all images from an SWF (shockwave flash) file. Returns a list
    of numpy arrays, or, if asNumpy is false, a list if PIL images.

    Limitation: only read the PNG encoded images (not the JPG encoded ones).

    """

    # Check whether it exists
    if not os.path.isfile(filename):
        raise IOError('File not found: '+str(filename))

    # Check PIL
    if (not asNumpy) and (PIL is None):
        raise RuntimeError("Need PIL to return as PIL images.")

    # Check Numpy
    if np is None:
        raise RuntimeError("Need Numpy to read SWF files.")

    # Init images
    images = []

    # Open file and read all
    fp = open(filename, 'rb')
    bb = fp.read()

    try:
        # Check opening tag
        tmp = bb[0:3].decode('ascii', 'ignore')
        if tmp.upper() == 'FWS':
            pass # ok
        elif tmp.upper() == 'CWS':
            # Decompress movie
            bb = bb[:8] + zlib.decompress(bb[8:])
        else:
            raise IOError('Not a valid SWF file: ' + str(filename))

        # Set filepointer at first tag (skipping framesize RECT and two uin16's
        i = 8
        nbits = bitsToInt(bb[i: i + 1], 5)  # skip FrameSize
        nbits = 5 + nbits * 4
        Lrect = nbits / 8.0
        if Lrect % 1:
            Lrect += 1
        Lrect = int(Lrect)
        i += Lrect+4

        # Iterate over the tags
        counter = 0
        while True:
            counter += 1

            # Get tag header
            head = bb[i:i+6]
            if not head:
                break # Done (we missed end tag)

            # Determine type and length
            T, L1, L2 = getTypeAndLen(head)
            if not L2:
                print('Invalid tag length, could not proceed')
                break
            #print(T, L2)

            # Read image if we can
            if T in [20, 36]:
                im = _readPixels(bb, i+6, T, L1)
                if im is not None:
                    images.append(im)
            elif T in [6, 21, 35, 90]:
                print('Ignoring JPEG image: cannot read JPEG.')
            else:
                pass # Not an image tag

            # Detect end tag
            if T == 0:
                break

            # Next tag!
            i += L2

    finally:
        fp.close()

    # Convert to normal PIL images if needed
    if not asNumpy:
        images2 = images
        images = []
        for im in images2:
            images.append(PIL.Image.fromarray(im))

    # Done
    return images



#!/usr/bin/env python

import numpy as np
import visvis as vv
from visvis import Point, Pointset
app = vv.use()

# create random points
a = np.random.normal(size=(1000,3))
pp = Pointset(a)
pp *= Point(2,5,1)

# prepare axes
a = vv.gca()
a.cameraType = '3d'
a.daspectAuto = False

# draw points
l = vv.plot(pp, ms='.', mc='r', mw='9', ls='', mew=0 )
l.alpha = 0.1

app.Run()
